from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import pytest

from random import randint, shuffle, sample

from mmgroup import mat24
from mmgroup import GCode, Cocode, PLoop, PLoopZ
from mmgroup import Octad, GcVector 
from mmgroup import AutPL
from mmgroup import Parity
from mmgroup.mat24 import MAT24_ORDER

from mmgroup import MM0
group = MM0

#####################################################################
# Test embedding of Cocode and AutPL into MMGroup
#####################################################################

@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [100])
def test_autploop(n_cases):
    for i in range(n_cases):
        autp = AutPL('r', 'r')
        g = group(autp)
        assert g ==  group([('d', autp.cocode), ('p',  autp.perm)])
        assert g ==  group('p',  autp)
        coc, p_num = autp.cocode, autp.perm_num
        if coc and p_num:
            assert g.as_tuples() == autp.as_tuples() 
            assert g.as_tuples() == [('d', coc), ('p', p_num)] 
        Coc = Cocode(coc)
        h1 = range(9, 18)
        h2 = autp.perm[9:18]
        assert g ==  group([('d', Coc), ('p',  zip(h1, h2))])    
        assert g ==  group([('d', Coc), ('p',  dict(zip(h1, h2)))])    


#####################################################################
# Test embedding of PLoop into MMGroup
#####################################################################

@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [100])
def test_xyz(n_cases):
    g1 = group()
    for i in range(n_cases):
        v = randint(0, 0x1fff)
        for tag in "xyz":
            g = group(tag, v) 
            assert g ==  group(tag, PLoop(v)) 
    group('x', v) * group('y', v) * group('z', v) == g1
 

#####################################################################
# Test multiplication/inversion in groups AutPL and MMGroup
#####################################################################


@pytest.mark.ploop
@pytest.mark.parametrize("n_cases", [100])
def test_group_op(n_cases):
    for i in range(n_cases):
        a1 = AutPL('r', 'r')
        c, p = randint(0, 0xfff), randint(0, MAT24_ORDER - 1)
        a2 = AutPL(c, p)
        assert group(a1) * (group(a2)) == group(a1 * a2)
        assert group(a1**-1) == group(a1)**-1





"""The module contains class ``BitMatrixMulFix``.

Class ``BitMatrixMulFix`` contains a directive that generates
code to multiply a fixed bit matrix with a variable bit matrix.

The C function ``mat24_perm_to_matrix`` uses this kind of matrix
multiplication to convert a permutation in the Mathieu group
to a bit matrix operating on a Golay code word.
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from random import randint
import types
import sys
import re
import os
from operator import __or__, __xor__
from functools import reduce

from mmgroup.bitfunctions import bitlen, bitweight, hibit, v2
from mmgroup.generate_c import UserDirective



#######################################################################
# Auxiliary functions for function ``make_addition_tree``
#######################################################################

def small_addition_tree(data):
    """Auxiliary function for function ``make_addition_tree``

    The function returns a dictionary with entries of shape 

    ``r:  (x, y)``

    for integers ``r, x, y`` such that ``r = x ^ y``, and  each of 
    the entries ``x, y`` is either of bit weight 1 or it is also
    contained as a key in the dictionry.

    The dictionary contains all entries of the list ``data`` of
    bit weight > 1 as keys.

    This function is highly recursive and should be used only if
    the bit weight of the unition of the entries of data
    is not too large (say, at most 8).
    """
    # Collect all unprocessed entries of array ``data`` in ``data1``
    # with bit weight atleast 2
    data1 = []   # list of entries yet to be processed
    for x in data:
        if bitweight(x) > 1 and not x in data1:
             data1.append(x)
    # Done if the list ``data1`` is empty.
    if len(data1) == 0:
        return {}
    # Sort list ``data1`` by descending bit weight 
    data1 = sorted(data1, key = lambda x: (-bitweight(x),-x))
    # Process first entry ``x`` of data1
    x = data1[0]
    # If list ``data1`` contains a word ``y`` with ``y = x & y``
    # then append an entry ``x: (y, x ^ y)`` to the output dict
    # and replace entry ``x`` in ``data1`` by ``x ^ y``.
    # Then recursively enter the result of
    # ``small_addition_tree(data1)`` to output dictionary ``dict``
    # adn return.
    for y in data1[1:]:
        if x & y == y:
            data1[0] = x ^ y
            dict = small_addition_tree(data1[:])
            dict[x] = (y, x ^ y)
            return dict
    # If the previous attempt has failed then find an entry ``y1``
    # (differnet from ``x```) in the list such that the bit 
    # weight of ``x & y1`` is maximal.
    w, y1 = 1, None
    for y in data1[1:]:
        w1 = bitweight(x & y)
        if w1 > w:
            w, y1 = w1, y
    # If a suitable ``y1`` of bit weight at least 2 has been found
    # then delete entry ``x`` from the list ``data1`` add the entry 
    # ``z`` with ``z = x & y1`` and also the entries ``x ^ z``, 
    # ``y1 ^ z``  to the list. The let the output dict consist of the 
    # entries ``x: (z, x ^ z),  y1:(z, y1 ^ z)`` and the recurssively
    # computed output ``small_addition_tree(data1)`` and return
    if not y1 is None: 
        data1[0] = z =  x & y1
        data1 += [x ^ z, y1 ^ z]
        dict = small_addition_tree(data1[:])
        dict[x] = (z, x ^ z)
        dict[y1] = (z, y1 ^ z)
        return dict
    # If no suitable ``y1`` of bit weight at least 2 has been found
    # then let ``y`` be the power of two with the higetst bit of ``x``
    # being set. Then append an entry ``x: (y, x ^ y)`` to the output 
    # and and replace entry ``x`` in ``data1`` by ``x ^ y``.
    # Then recursively enter the result of
    # ``small_addition_tree(data1)`` to output dictionary ``dict``
    # and return.
    y =  1 << hibit(x)
    data1[0] ^= y
    dict = small_addition_tree(data1[:])
    dict[x] = (y, x^y)  
    return dict  


#######################################################################
# Function ``make_addition_tree``
#######################################################################


def make_addition_tree(data, singleton = False, granularity = 8):
    """Create a direct addition tree for a set of integers

    The function returns a pair ``(dict, stage)``. Here 
    dictionary ``dict`` has entries of shape ``r:  (x, y)``
    for integers ``r, x, y`` such that ``r = x ^ y``, and  each of 
    the entries ``x, y`` is either of bit weight 1 or it is also
    contained as a key in the dictionry.

    The dictionary contains all entries of the list ``data`` of
    bit weight > 1 as keys.

    This function calls ``small_addition_tree`` (possibly several 
    times) with a list of data, such that the bit weight of 
    the union of these data is at most equal to ``granularity``.

    if parameter ``singleton`` is True then all bit positions 
    set in any entry of ``date`` are entered into the dictionary 
    as key of bit weight 1, with is value being an empty tuple.

    Dictionary ``shape`` has the same set of keys as dictionary
    ``dict``. For each key the value in that dictionary indicates 
    a **stage** at which this entry is computed.
    """
    dict = {}
    stage = {}
    if len(data) == 0 or max(map(bitweight, data)) <= 1:
        return dict, stage
    all_bits =  reduce(__or__, data, 0)
    maxbit = bitlen(all_bits)
    st = 0
    for i in range(0, maxbit, granularity):
        mask1 = (1 << (i + granularity)) -  (1 << i)
        mask  = (1 << (i + granularity)) -  1
        a = [] 
        for x in data:
            xm1 = x & mask1
            xm  = x & mask 
            xm0 = xm1 ^ xm
            if bitweight(xm0) and bitweight(xm1): 
                dict[xm] = (xm1, xm0)
                stage[xm] = st+1
            
            if bitweight(xm1) > 1 and not xm1 in a:
                a.append(xm1)
        dict1 =  small_addition_tree(a)
        for d in dict1.keys():
            dict[d] =  dict1[d]
            stage[d] = st
        st += 2
    # Check addition tree in ``dict``
    #show_addition_tree(dict) 
    for x in (dict):
        for y in dict[x]:
            if bitweight(y) > 1: assert y in dict.keys(), (hex(x),hex(y))
    # Add singeltons to ``dict`` if requested
    if singleton:
        while all_bits:
            entry = all_bits & -all_bits
            dict[entry] = ()
            all_bits &= ~entry

    return dict, stage

#######################################################################
# Auxiliary functions for function ``make_addition_operations``
#######################################################################
      
    
def calc_fanout(data, dict):
    big = 2*len(dict) + 2
    fanout = {}
    for x in dict.keys():
        fanout[x] = big if x in data else 0
    for x in dict.keys():
        for y in dict[x]:
            if y in dict.keys():
                 fanout[y] += 1
    return fanout
     


class register_set(object):
    def __init__(self):
        self.free = []
        self.n = 0
    def alloc(self):
        if not len(self.free):
            self.free.append(self.n)
            self.n += 1
        return self.free.pop()
    def dealloc(self, n):
        assert 0 <= n < self.n
        self.free.append(n)
    def alloc_index(self,n):
        while self.n <= n:
           self.free.append(self.n)
           self.n += 1
        i = self.free.index(n)
        del self.free[i]
        return n
                   
    


def reassign_registers(data, registers):
    mapping = {}
    free_registers =  register_set()
    for i,x in enumerate(data):
        if x in registers.keys() and not registers[x] in mapping.keys():
             mapping[registers[x]] = free_registers.alloc_index(i)
    for r in registers.values():
        if not r in mapping.keys():
             mapping[r] = free_registers.alloc()
    new_registers = {}
    for x in registers.keys():
        new_registers[x] = mapping[registers[x]]
    return new_registers

       



def make_add_operations(data, singleton = False, granularity = 4):
    calc,stage =  make_addition_tree(data, singleton, granularity)
    fanout =  calc_fanout(data, calc)   
    done = set()
    big = 2*len(calc) + 2
    registers = {}
    ops = []  
    allocated_regs = register_set() 

    for x, d in calc.items():
        if bitweight(x) <= 1:
            registers[x] = allocated_regs.alloc()
            done.add(x) 
            ops.append((x, x))
        for y in d:
            if bitweight(y) <= 1:
                done.add(y)
    remain = set(calc.keys()) - done


    while len(remain):
        fo, y1 = [3, big*big]  , None
        x = None
        for y in remain:
             doable = calc[y][0] in done and calc[y][1] in done   
             fo_new = [2, stage[y]] 
             for predec in  calc[y]:
                 if predec in calc.keys() and fanout[predec] == 1:
                     fo_new[0] -= 1
             if doable and fo_new < fo:
                 fo, y1 = fo_new, y                

        for predec in calc[y1]:
            if predec in calc.keys():
                 fanout[predec] -= 1
                 if fanout[predec] == 0:
                     allocated_regs.dealloc(registers[predec])
        registers[y1] =  allocated_regs.alloc()
        ops.append((y1,)+calc[y1])
        remain.remove(y1)
        done.add(y1)

    registers = reassign_registers(data, registers)
    return  ops, registers


def copy_operation(data, ops, registers):
    copies = []
    zeros = []
    for i, x in enumerate(data):
        if not x in registers.keys() or registers[x] != i:
           if x:
                copies.append( (i, x, x) )
           else:
                zeros.append(i)
    for i in zeros:
        copies.append( (i, 0, 0) )
    return copies


#######################################################################
# Function ``make_addition_operations``
#######################################################################


def make_addition_operations(m, singleton = False, granularity = 4):
    """Straight line program multiplying fixed with vatiable bit matrix

    The function returns a staight-line programm for multiplying
    a fixed bit matrix ``m`` with a variable bit matrix ``a``. Here
    a bit matrix is encoded as an array (or a list) of integers as in 
    module ``mmgroup.bitfunctions``. Matrix ``m`` is the left factor 
    of the matrix product. Multiplication with matrix ``m`` from
    the left corresponds to a sequence of row operations applied to
    matrix ``a``.

    The straight line program for computing ``m * a`` is returned
    as a array of tuples ``(register, value, op1 [, op2])``, where
    each tuple encodes a binary or a copy operation. If entry 
    ``op2`` is not present then we have a copy operation. In a 
    nutshell this tuple encodes the operation:

    Compute ``(value * a) = (op1 * a) + (op2 * a)`` and store the 
    result in register ``register``.

    ``register`` is the number of register where to store the result 
    of the operation. Register numbers start with 0. Eventually,
    ``register[i]`` will contain the ``i``-th row of the output
    matrix. That number of rows of the result ``m * a`` is the 
    maximum of the bit lengths of the entries of input entry ``m``; 
    so the last row of the output matrix is the highest column 
    of ``m`` with nonzero  entries. More registers may be required 
    for storing intermediate results.
   
    Operands and results of an operation are bit vectors corresponding
    to rows of bit matrices. Here the value ``v`` of an entry 
    ``value``, ``op1``, or ``op2`` is an integer encoding a bit vector 
    ``v``, such  that this entry is equal to the bit vector ``v * a``. 
    If ``op2`` is present then the operation encoded by this tuple is 
    the addition of these two operands (as bit vectors); and we 
    have ``value = op1 ^ op2``. Otherwise we have a copy operation
    an then we have  ``value = op1``.

    In case ``v = 0`` the corrresponding bit vector is zero; and if
    ``v`` has bit weight 1 then the corrsponding bit vector is a row
    of matrix ``a``.  A bit vector corrsponding to any other value 
    of ``v`` must be (and will be) computed in a register.

    So the user may generate e.g. C code for computing ``m * a``
    from ``a`` by using the output of this function. Here the user
    should maintain a dictionary that maps a value ``v``
    (corresponding to a bit vector ``v * a``) to the number of
    the register containing that bit vector. This dictionary
    is required for locating non-trivial operands in the 
    register set.

    If parameter ``singleton`` is True then all rows of the input
    matrix ``a`` are loaded into the registers before any binary
    operations will be done. 

    Parameter ``granularity`` controls the construction of the
    addition tree as in function ``make_addition_tree``.
    """
    ops, regs = make_add_operations(m, singleton)
    copies = copy_operation(m, ops, regs)
    ops_out = []
    for op in ops:
        d = (regs[op[0]],) + op
        ops_out.append(d)
    return ops_out + copies



#######################################################################
# Testing function ``make_addition_operations``
#######################################################################



def check_addition_operations(m, ops, singleton = False):
    """Check result of function ``make_addition_operations``

    Here parameters ``m`` and ``singletons`` are inputs to
    function ``make_addition_operations`` and parameter ``ops``
    is the corresponding return value of that function.

    Then this function ``check_addition_operations`` performs a
    straight-line program computing ``m * a`` for the unit  matrix 
    ``a`` based on parameter ``ops``. It raises ValueError if the
    result of that straight-line program differs from ``m``.

    If parameter ``singleton`` is True than we also check the
    condition for this case as stated in the documentation of
    function ``make_addition_operations``.    
    """
    num_regs = max([op[0] for op in ops]) + 1
    results = {}
    registers = {}
    
    def check_operand(value):
        in_registers = value in registers
        if  bitweight(value) > 1:
            assert in_registers
        if in_registers:
            assert results[registers[value]] == value
        return in_registers

    for op in ops:
        reg, value, operands = op[0], op[1], op[2:]
        assert len(operands) > 0
        in_registers = min([check_operand(value) for value in operands])
        if len(operands) > 1 and singleton:
            assert in_registers
        assert value == reduce(__xor__, operands)
        results[reg] = value
        registers[value] = reg

    for i, x in enumerate(m):
        assert results[i] == x
    return


#######################################################################
# Display result of function ``make_addition_operations``
#######################################################################


def display_addition_operations(ops):
    """Display result of function ``make_addition_operations``

    Let ``ops`` be a result of a call to function 
    ``make_addition_operations``. Then we display the straight-
    line program corrsponding to that result in a readable form.
    """
    print("Straight-line program generated by make_addition_operations") 

    registers = {}
    
    def display_operand(value):
        if value in registers:
            return 'o[%d]' % registers[value]
        if value == 0:
            return "0"
        if  bitweight(value) == 1:
            return "i[%d]" %  v2(value)
        raise ValueError("Illegal operand in straight line program")

    for op in ops:
        reg, value, operands = op[0], op[1], op[2:]
        op_str = " ^ ".join([display_operand(s) for s in operands])
        assign_str = "%-5s = %s" % ("o[%d]"  % reg, op_str)
        print("%-30s // 0x%08x" % (assign_str, value))
        registers[value] = reg


#show_addition_tree = display_addition_operation

#######################################################################
# Class ``BitMatrixMulFix`` wrapping ``make_addition_operations``
#######################################################################


class BitMatrixMulFix(object):
    """Models multiplication of a fixed by a variable bit matrix.


    The left factor is a fixed bit matrix. Bit matrices are encoded
    as arrays of integers as in module bitfunctions.

    The purpose of this class is to generate C code for perfroming
    this multiplication.
    """

    def __init__(self, granularity = 4):
        """Here 'fixed_matrix' is the fixed left bit matrix factor. 

        'granularity' means something like the number of columns of
        the fixed matrix that are processed in one stage. The default
        appears to be appropriate for m times n matrices,
        12 <= m, n <= 24. 
        """
        self.granularity = granularity

    def set_fixed_matrix(self, fixed_matrix, singleton = False):
        """Here 'fixed_matrix' is the fixed left bit matrix factor. 

        Here we use some magic (vulgo: poorly documented) method
        to obtain a straight-line program calculating the
        non-trivial lines of the matrix product via xor operations.

        If you just want to understand the result (not the magic),
        it is best to check member function selftest.

        """
        self.data = list(fixed_matrix)
        self.singleton = singleton
        self.nrows =  len(self.data)  
        self.ncols = bitlen(reduce(__or__, self.data, 0))
        self.ops = make_addition_operations(
            fixed_matrix, singleton, self.granularity
        )
        self.n_registers = max([op[0] for op in self.ops]) + 1
        self.selftest()



    def set_names(self, type_name, input_name, output_name):
        self.type_name = type_name
        self.input_name = input_name
        self.output_name = output_name
        self.tmp_name = "_r"
        i = 0
        while self.tmp_name in [self.input_name, self.output_name]:
            self.tmp_name = "_r"+str(i)
            i += 1

    def array_entry(self, array_name, index):
        return array_name + "[" + str(index) + "]"

    def register_name(self, reg):
        if reg >= self.nrows:
            return self.array_entry(self.tmp_name, reg - self.nrows)
        else: 
            return self.array_entry(self.output_name, reg)

    def operand_name(self, x, registers):
        if x in registers:
            return self.register_name(registers[x])
        if x == 0:
            return "0"
        r = v2(x)
        assert x == 1 << r
        return "(%s)(%s)" % (self.type_name,
                                self.array_entry(self.input_name, r))
 
    def make_c_calculation(self):
        s = ""
        registers = {}
        for op in self.ops:
            reg, value, operations = op[0], op[1], op[2:]
            res_name = self.register_name(reg)
            operand_names = [
                self.operand_name(x, registers) for x in operations]
            registers[value] = reg
            operation = " ^ ".join(operand_names)
            s +=  " %s = %s;\n" % (res_name, operation)
        return s;

    def ops(self):
        return self.ops

    def make_c_comment(self):
        s = """GF(2) bit matrix multiplication as a sequence of XOR operations:
  %s  = _MFIX_ * %s  ,
with bit matrices represented as arrays of unsigned integers,
each integer representing a row of the matrix. 
Here _MFIX_ is the fixed bit matrix  {\n""" % (
   self.output_name, self.input_name)

        w = max(1, (self.ncols + 3) >> 2)
        l = (60) // (w+3)
        for i, d in enumerate(self.data):
            if i%l == 0:  s += " "
            s += "0x%0*x" % (w,d)
            if i ==  len(self.data)-1: 
                s += "\n}"
            elif i%l == l-1: 
                s += ",\n"
            else: 
                 s += ","
        lines = [  "// " + y for y in s.split("\n")] 
        return "\n".join(lines)   + "\n"

    def make_c_program(self):
        s = "\n" + self. make_c_comment() +  "{" + "\n"
        if self.n_registers > self.nrows:
            s +=  " %s %s[%d];\n" % (self.type_name,self.tmp_name,
                 self.n_registers - self.nrows)
        s += self.make_c_calculation()
        s += " // %d operations generated\n" % len(self.ops)
        s +=  "}" + "\n"
        return "\n" + s + "\n"
                                
 
    def generate_c_bitmatmul(self, table, type, input_var, output_var): 
        self.set_fixed_matrix(table)
        self.set_names(type, input_var, output_var)
        return self.make_c_program()

    def selftest(self):
        check_addition_operations(self.data, self.ops, self.singleton)
 

    def tables(self):
        """This class generates no tables"""
        return {}

    def directives(self):
        """Code generation directive has name "BITMATMUL". """
        return {
           "BITMATMUL" : UserDirective(self.generate_c_bitmatmul, "psss"),
        }
    
   


from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import random
from random import randint
import types
import sys
import re
import os
from operator import __or__, __xor__
import subprocess

import pytest
 
from mmgroup.tests.tempdir import make_temp_dir, kill_temp_dir
from mmgroup.tests.compile import compile_testprogramm
from mmgroup.dev.mat24.make_mul_transp import BitMatrixMulTransp
#from mmgroup.dev.mat24.make_mul_transp import generate_c_mul_transp
  

#########################################################################
######### Tests #########################################################
#########################################################################



@pytest.mark.slow
@pytest.mark.mat24
@pytest.mark.compiler
def test_make_mul_transp():
    """Generate, compile and test a C program"""
    tmp_dir =  make_temp_dir()
    FILE_NAME = "tmp_test_mul_transp"
    NAME = os.path.join(tmp_dir, FILE_NAME)
    C_NAME, EXE_NAME = NAME + ".c", NAME + ".exe"
  
    def make_testprogram(lv, lm, n=1):
        generate = BitMatrixMulTransp().generate_c_mul_transp
        f = open(C_NAME, "wt")
        print(r"""#include <stdio.h>
#include <stdint.h>
int main(int argc, char **argv)
{{
    uint_fast32_t i, v;
    static uint_fast32_t a[1000];
    sscanf(argv[1], "%ld", &v);
    for (i=0; i < argc-2; ++i) sscanf(argv[i+2], "%ld", a+i);
    {0}
    printf("%ld\n",v);
    return 0;
}}
""".format(generate("v", "a", lv, lm, n)), file = f)
        f.close()
        #subprocess.check_output(["gcc", C_NAME, "-o", EXE_NAME])
        compile_testprogramm([C_NAME], EXE_NAME)
        checker = BitMatrixMulTransp()
        checker.set_matrix(lv, lm, n)
        return checker

    def run_testprogram(v, m):
        data = list(map(str, [v]+m))  
        res =  subprocess.check_output([EXE_NAME] + data)
        return int(res)  

    def test_testprogram(checker , v, m):
        res =  run_testprogram(v, m)
        ref = checker.compute(v,m+[0]*64)
        assert ref == ref, (hex, ref)

    def del_testprogram():
        for f in [C_NAME, EXE_NAME]:
            os.remove(f)

    test_data_dict = {
        (4,4,1,0): [ (3, [2,3, 1, 1]) , ]
    }

    def test_data(lv, lm, n, n_tests=10):
        """yield  n_tests test cases (v, m) for given lv, lm, m0"""
        try:
            d =  test_data_dict[(lv, lm, n, m0)]
            for v, m in d: yield v, m
        except:
            pass
        for i in range(n_tests):
            v = random.randint(0, 2**(n*lv)-1)
            m = [] 
            for i in range(lm):
                 m.append(random.randint(0, 2**(n*lv)-1))
            yield v, m

    def test_programs():
        """yield cases (lv, lm, n) for a making a test program"""
        yield 4, 4, 1
        yield 16, 12, 2

    

       
    for lv, lm, n in test_programs():
        print("Make C program for case lv=%d, lm=%d, n=%d" % (
                 lv,lm,n))
        checker = make_testprogram(lv, lm, n)  
        print ("C program has been made, starting tests..")
        for v, m in test_data(lv, lm, n):
              test_testprogram(checker, v, m)
        del_testprogram()
        print ("passed")

    kill_temp_dir(tmp_dir)





r"""

Module ``generators`` contains the definition of the generators of the 
monster, so that they may be used in C files. It also contains support 
for the subgroups :math:`N_{0}` of structure 
:math:`2^{2+11+2\cdot11}.(\mbox{Sym}_3 \times M_{24})` and 
:math:`G_{x0}` of structure :math:`2^{1+24}.\mbox{Co}_1` of the 
monster, as described in :cite:`Con85` and :cite:`Seysen20`. Here
:math:`M{24}` is the Mathieu group acting on 24 elements, and 
:math:`\mbox{Co}_1` is the automorphism group of the 24-dimensional
Leech lattice modulo 2.

Here we fully support the computation in the subgroup :math:`N_{0}` 
based on the generators defined in this module, so that a word in the 
generators of :math:`N_{0}` can easily be reduced to a standard form. 

We also support the operation of the group :math:`G_{x0}` on the 
Leech lattice mod 2 and mod 3 (in some cases up to sign only). For a 
full support of the subgroup :math:`G_{x0}` we also have to compute 
in a Clifford group, which is implemented in module ``clifford12``.

Our set of generators of the monster group is defined in section
:ref:`mmgroup-label`. The C implementation of this set of generators
is defined in section *Header file mmgroup_generators.h*.


The intersection :math:`N_{0} \cap G_{x0}` is a group :math:`N_{x0}` 
of structure :math:`2^{1+24}.2^{11}.M_{24}`. The group :math:`M_{24}` 
(and, to some extent, also the group :math:`N_{x0}`) is 
supported by the C functions in file ``mat24_functions.c``. 

The Leech lattice and the extraspecial group :math:`Q_{x0}`
-----------------------------------------------------------

Let :math:`Q_{x0}` the normal subgroup of :math:`G_{x0}` of structure
:math:`2^{1+24}`. Then :math:`Q_{x0}` is an extraspecial 2 group and
also a normal subgroup of :math:`G_{x0}`. Let :math:`\Lambda` be the
Leech lattice. The quotient of :math:`Q_{x0}` by its center 
:math:`\{\pm1\}` is isomorphic to  :math:`\Lambda/2 \Lambda`, which
is the Leech lattice modulo 2.


For :math:`e_i \in Q_{x0}` let :math:`\tilde{e}_i` be the vector in 
the Leech lattice (mod 2) corresponding to :math:`\pm e_i`. Then
:math:`e_1^2 = (-1)^s` for 
:math:`s = \langle\tilde{e}_1, \tilde{e}_1 \rangle /2`,
where :math:`\langle.,.\rangle` is the scalar product in the
Leech lattice. For the commutator :math:`[e_1, e_2]` we have 
:math:`[e_1, e_2] = (-1)^t`, 
:math:`t = \langle \tilde{e}_1, \tilde{e}_2 \rangle`.


Leech lattice encoding of the elements of :math:`Q_{x0}`
--------------------------------------------------------

An element of  of :math:`Q_{x0}` can be written uniquely as 
a product :math:`x_d \cdot x_\delta`, 
:math:`d \in \mathcal{P} , \delta \in \mathcal{C}^*`, see
:cite:`Seysen20`, section 5. Here :math:`\mathcal{P}` is the 
Parker loop and :math:`\mathcal{C}^*` is the Golay cocode.
We encode the element 
:math:`x_d \cdot x_\delta` of :math:`Q_{x0}` as an integer 
:math:`x`  as follows:

.. math::

       x = 2^{12} \cdot d \oplus (\delta \oplus \theta(d)) \, .

Here elements of the Parker loop and elements of the cocode 
are encoded as integers as in section :ref:`parker-loop-label`
and :ref:`golay-label`.
:math:`\theta` is the cocycle given in section 
:ref:`basis-golay-label`, and ':math:`\oplus`' means bitwise
addition modulo 2. Note that a  Parker loop element is 13 bits 
long (with the most significant bit denoting the sign) and that 
a cocode element is 12 bits long.

From this representation of :math:`Q_{x0}` we obtain a 
representation of a vector in the Leech lattice modulo 2 by 
dropping sign bit, i.e. the most significant bit at position 24. 
A vector addition in the Leech lattice modulo 2 can be done by 
applying the XOR operator ``^`` to the integers representing 
the vectors, ignoring the sign bit.
 
Special elements of the group :math:`Q_{x0}`
--------------------------------------------

We write :math:`\Omega` for the positive element of the Parker 
loop such that :math:`\tilde{\Omega}` is the Golay code word
:math:`(1,\ldots,1)` as in :cite:`Con85` and :cite:`Seysen20`.
In this specifiction we also write :math:`\Omega` for the 
element :math:`x_{\Omega}` of :math:`Q_{x0}` and for the element 
:math:`\tilde{x}_{\Omega}` of the Leech lattice modulo 2 if the 
domain of :math:`\Omega` is clear from the context. Then
:math:`\Omega` has Leech lattice encoding ``0x800000`` in
our chosen basis of the Golay code; and the element :math:`\Omega` 
of  :math:`\Lambda/2 \Lambda` corresponds to the standard
coordinate frame of the real Leech lattice.

For fast computations in the monster group it is vital to compute
in the centralizer of a certain short element :math:`x_{\beta}`
of :math:`Q_{x0}`, where :math:`\beta` is an even coloured element 
of the Golay cocode, as described in :cite:`Seysen20`. Here we 
choose the cocode element :math:`\beta` corresponding to the
element :math:`(0,0,1,1,0,\ldots,0)` of  :math:`\mbox{GF}_2^{24}`.
Then the centralizer of :math:`x_{\beta}` is isomorphic to the
a double cover baby monster group and contains the generators 
:math:`\tau` and  :math:`\xi` of the monster. We also write  
:math:`\beta` for the element :math:`x_{\beta}` of :math:`Q_{x0}` 
and for the element :math:`\tilde{x}_{\beta}` of 
:math:`\Lambda/2 \Lambda` in the same way as for the element 
:math:`\Omega`.  Then :math:`\beta` has Leech lattice 
encoding ``0x200``  in our chosen basis.

Module ``gen_leech_reduce.c`` contains functions for rotating 
arbitrary type-4 vectors in :math:`\Lambda/2 \Lambda` to 
:math:`\Omega` and for  rotating arbitrary type-2 vectors in 
:math:`\Lambda/2 \Lambda` to :math:`\beta`. 

Computations in the Leech lattice modulo 3
------------------------------------------

For the construction of the subgroup :math:`G_{x0}` of the monster
we also require the automorphism group :math:`\mbox{Co}_0` of the
**real** Leech lattice, as decribed in :cite:`Con85`. That group has 
a faithful representation as an automophism group of 
:math:`\Lambda/3 \Lambda`, but not of :math:`\Lambda/2 \Lambda`. 
Module ``gen_leech3.c`` provides functions for computing in the 
Leech Lattice modulo 3.

Note that in :cite:`Seysen20` the operation of the generators 
:math:`x_d, y_d, x_\delta, x_\pi, \xi` of :math:`G_{x0}` is
also defined on the real Leech lattice.
"""
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from mmgroup.bitfunctions import bw24, lmap

import random
import numpy
from numpy import array, zeros, uint8, uint16, uint32

from mmgroup.dev.mat24.mat24_ref import Mat24


#######################################################################
# Auxiliary functions 
#######################################################################


def w2_gamma(v):
    """Implement function gamma() in [Seysen20], section 3.3.

    Given a Golay code vector v  in 'gcode' representation, see
    module mat24.py, we return a pair w2, c. Here c is equal to
    the element gamma(v) of the Golay cocode, with gamma() as in 
    [Seysen20], section 3.3. ,c is returned in 'cocode' representation, 
    see module mat24.py. We also return the bit w2 = w2(c), with w2()
    as defined in [Seysen20], section 3.3.   
    """
    x = Mat24.gcode_to_vect(v)
    x1 = sum( (x >> i) & 0x111111 for i in range(1,4) )
    x1 = (x1 >> 1) & 0x111111
    w2 = bw24(x1)
    w2 = ((w2 * (w2 - 1)) >> 1) & 1
    c = Mat24.vect_to_cocode(x1)
    return w2, c



def compress_gray(x):
    """Extract the gray part of a Colay code or cocode word

    The function returns the 'gray' part of a Golay code or cocode
    as a 6-bit number. It drops the 'colored' part of the word.

    A 12-bit Golay code or cocode word can be expressed as a sum of
    a 6-bit 'gray' and a 6-bit 'colored' word, see [Seysen20], 
    section 2.2.  

    A Golay code word x must be given in 'gcode' and a cocode word 
    must be given in 'cocode' representation. 
    """
    return (x & 0x0f) + ((x >> 6) & 0x30)

def expand_gray(x):
    """Reverse the effect of function compress_gray()

    Given a compressed 'gray' Golay code or cocode word, the function 
    returns the original 'gray' word in code or cocode representation.
    """
    return (x & 0x0f) + ((x & 0x30) << 6) 



#######################################################################
# Class Mat24Xi
#######################################################################



class GenXi(object):
    """Operation of xi of a group G_x0 on the normal subgroup Q_x0.

    A subgroup G_x0 of shape 2**{1+24}.Co_1 of the Monster group is 
    used for the construction of the monster group M in [Con85]. G_x0 
    is generated by a subgroup N_x0 of shape 2**{1+24}.2**{11}.Mat24, 
    and an additional element xi of order 3, see [Seysen20], section 9. 
    Mat24 is the Mathieu group acting on 24 elements. Mat24 (and also, 
    to some extent, N_x0) is supported by class mat24.Mat24. 

    G_x0 has an extraspecial normal subgroup Q_x0 of shape 2**{1+24},
    see [Con85], [Seysen20]. Class Mat24Xi focusses on the operation of 
    xi and x**2 on Q_x0 by conjugation.

    This class provides a table-based implementation of the operation
    of xi and xi**2, which is based on [Seysen20], Lemma 9.5. The class 
    also provides some auxiliary functions used for that operation 
    defined in [Seysen20].
    
    The so-called short elements of Q_x0 (see [Con85], sect.7) play an 
    important role as basis vectors in the representation of the 
    monster in [Con85] and in this project. Class Mat24Xi also contains 
    functions for mapping short vectors in Q_x0 to basis vectors of 
    that representation and vice versa. The basis for our representa-
    tion of the monster group, is described in module mm_aux.c.

    A faster C implementation of this class is given in class
    mat24fast.Mat24Xi      

    References
    ----------
    see file refereces.txt
    """
    tab_g_gray = numpy.zeros(64, dtype = numpy.uint8)
    tab_g_cocode = numpy.zeros(64, dtype = numpy.uint8)
    #tab_g_col = numpy.zeros(64, dtype = numpy.uint8)

    tables = {
            "GenXi_module_doc" : globals()["__doc__"],
            "GenXi_g_gray_table" : tab_g_gray,
            "GenXi_g_cocode_table" : tab_g_cocode,
    }

    directives = {}

    for x in range(64):
        w2, c = w2_gamma(expand_gray(x))
        cx = compress_gray(c)
        w2x = w2 << 7 
        tab_g_gray[x] = w2x + cx 
        tab_g_cocode[cx] = w2x + x


    

    @classmethod
    def gen_xi_g_gray(cls, v):
        """Implement function gamma() in [Seysen20], section 3.3.

        Given a Golay code vector v  in 'gcode' representation, see
        module mat24.py, we return the element gamma(v) of the Golay 
        cocode, with gamma() as in [Seysen20], section 3.3. c is 
        returned in 'cocode' representation, see module mat24.py.
        """
        return expand_gray(cls.tab_g_gray[compress_gray(v)])
  
    @classmethod
    def gen_xi_w2_gray(cls, v):
        """Implement function w2() in [Seysen20], section 3.3.

        Given a Golay code vector v  in 'gcode' representation, see
        module mat24.py, we return the bit w2 = w2(c), with w2() as
        defined in [Seysen20], section 3.3.   
        """
        return cls.tab_g_gray[compress_gray(v)] >> 7

    @classmethod
    def gen_xi_g_cocode(cls, c):
        """Inverse of method xi_g_gray(v)

        Given a cocode vector c in in 'cocode' representation, see 
        module mat24.py, the function returns the unique gray Golay
        code vector v such that cls.gen_xi_w2_gray(v) is the gray 
        part of c. 
        """
        return expand_gray(cls.tab_g_cocode[compress_gray(c)])
  
    @classmethod
    def gen_xi_w2_cocode(cls, c):
        """Implement function w2() in [Seysen20], section 3.3.

        Given a cocode vector c in in 'cocode' representation, see 
        module mat24.py, the function returns the bit w2 = w2(c), 
        with w2() as defined in [Seysen20], section 3.3.   
        """
        return cls.tab_g_cocode[compress_gray(c)] >> 7

    @classmethod
    def gen_xi_op_xi(cls, x, exp):
        """Operation of  xi**exp  on the element x of the group Q_x0.

        The function returns the element

            xi**(-exp)  *  x  *  xi**exp

        of the group Q_x0. The element x of Q_x0 as must be given in
        Leech lattice encoding, as described in the header of this 
        module.

        The returned result is is coded in the same way. This 
        function implements the formula in [Seysen20], Lemma 9.5.
        """
        exp %= 3
        if (exp == 0): 
            return x
        scal = bw24((x >> 12) &  x & 0xc0f) & 1  
        x ^= scal << 24 # xor scalar product to sign

        tv =  cls.tab_g_gray[compress_gray(x >> 12)] 
        w2v, gv = tv >> 7, expand_gray(tv)
        tc =  cls.tab_g_cocode[compress_gray(x)] 
        w2c, gc = tc >> 7, expand_gray(tc)
        if (exp == 1):
            x &= ~0xc0f000  # kill gray code part
            x ^= w2c << 24  # xor w2(cocode) to sign
        else:
            x &= ~0xc0f     # kill gray cocode part
            x ^= w2v << 24  # xor w2(code) to sign
        x ^= gv         # xor g(code) to cocode 
        x ^= gc << 12   # xor g(cocode) to code
        return x



        
    @staticmethod
    def gen_xi_short_to_leech(x1):
        """Convert short vector to Leech lattice encoding.

        Both, Leech lattice and short vector encoding of a short vector 
        in Q_x are decribed in the header of this module. The function 
        returns the Leech lattice encoding of element x1 given in short
        vector encoding. 

        The function returns 0 for an illegal input x1. 
        """   
        box, sign, code = x1 >> 16, (x1 >> 15) & 1, x1 & 0x7fff
        octad = 0xffff
        if box == 1:
            if code < 1536:  # this is 2 * 24 * 32 
                gcode = code >= 768
                if gcode:
                    code -= 768
                gcode <<= 11
                i, j = code >> 5, code & 31
                cocode = Mat24.vect_to_cocode((1 << i) ^ (1 << j))
                if cocode == 0 or cocode & 0x800:
                    return 0
            elif code < 2496: # this is 2 * 24 * 32 + 15 * 64 
                octad = code - 1536
            else:
                return 0
        elif box == 2:
            if code >= 23040:  # this is  360 * 64     
                return 0   
            octad = code + 960  # this is 15 * 64
        elif box == 3:
            if code >= 24576:  # this is  384 * 64     
                return 0   
            octad = code + 24000  # this is (15 + 360) * 64
        elif box < 6:
            code += (box - 4) << 15
            cocode = Mat24.vect_to_cocode(1 << (x1 & 31))
            if cocode  == 0:
                return 0
            gcode = (code >> 5) & 0x7ff
            w =  Mat24.gcode_weight(gcode) ^ Mat24.scalar_prod(gcode, cocode) 
            gcode ^= (w & 1) << 11
        else:
            return 0  
        if octad < 48756:  # this is  759 * 64 
            cc = octad & 0x3f
            w = Mat24.bw24(cc) 
            cc = (cc << 1) ^ (w & 1)
            w += w & 1
            gcode = Mat24.octad_to_gcode(octad >> 6)
            gcodev = Mat24.gcode_to_vect(gcode)
            cocode = Mat24.vect_to_cocode(Mat24.spread_b24(cc, gcodev))
            gcode ^=  ((w >> 1) & 1) << 11
        cocode ^= Mat24.ploop_theta(gcode)
        return (sign << 24) | (gcode << 12) | cocode

    @staticmethod
    def gen_xi_leech_to_short(x1):   
        """Convert Leech lattice to short vector encoding.

        Both, Leech lattice and short vector encoding of a short vector 
        in Q_x are decribed in the header of this module. The function 
        returns the short vector encoding of element x1 given in Leech 
        lattice encoding. 

        The function returns 0 if the vector x1 is not short. 
        """   
        sign = (x1 >> 24) & 1
        x1 ^= Mat24.ploop_theta(x1 >> 12)
        gcodev = Mat24.gcode_to_vect(x1 >> 12) 
        cocodev = Mat24.cocode_syndrome(x1, 
            min(23, Mat24.lsbit24(gcodev))
        )
        w = Mat24.gcode_weight(x1 >> 12)
        if x1 & 0x800:
            if (Mat24.bw24(cocodev) > 1 or
                Mat24.scalar_prod(x1 >> 12, x1) !=  (w & 1)):
                return 0
            y = Mat24.lsbit24(cocodev)
            code = (x1 & 0x7ff000) >> 7 | y  
            box = 4 + (code >> 15)
            code &= 0x7fff
        else: 
            if w == 3:
                return 0
            elif w in [2,4]:
                if w == 4:
                    gcodev ^= 0xffffff
                    x1 ^= 0x800000
                w_bad = (Mat24.bw24(cocodev) ^ 2 ^ w) & 3
                if w_bad or (cocodev & gcodev) != cocodev:
                    return 0
                c =  Mat24.extract_b24(cocodev, gcodev)
                if (c & 0x80):
                     c ^=  0xff 
                y1 = Mat24.gcode_to_octad(x1 >> 12)
                code = (y1 << 6) | (c >> 1)
                if code >= 24000: # this is (15 + 360) * 64
                    code -= 24000
                    box = 3
                elif code >= 960: # this is 15 * 64
                    code -= 960
                    box = 2
                else:
                    code += 1536
                    box = 1
            else:
                y1 = Mat24.lsbit24(cocodev) 
                cocodev ^= 1 << y1  
                y2 = Mat24.lsbit24(cocodev) 
                if cocodev != (1 << y2) or y1 >= 24:
                    return 0
                code = 384 * (w & 2) + 32 * y2 + y1
                box = 1
        return  (box << 16) | (sign << 15) | code
  

    @classmethod
    def gen_xi_op_xi_short(cls, x, exp):
        """Operation of  xi**exp  on the element x of the group Q_x0.

        The function returns the element

            xi**(-exp)  *  x  *  xi**exp

        of the group Q_x0. The element x of Q_x0 as must be given in
        short vector encoding, as described in the header of this 
        module.

        The returned result is is coded in the same way. 
        """
        y = cls.gen_xi_short_to_leech(x)
        if (y == 0): return x
        y = cls.gen_xi_op_xi(y, exp)
        if (y == 0): return x
        y =  cls.gen_xi_leech_to_short(y)
        if (y == 0): return x
        return y

        
    @classmethod
    def make_table(cls, u_box, u_exp):
        assert 1 <= u_box <= 5 and 1 <= u_exp <= 2
        t_size = [0, 2496, 23040, 24576, 32768, 32768]
        a = numpy.zeros(32768, dtype = numpy.uint16) 
        length = t_size[u_box] 
        u_box <<= 16;
        for i  in range(length):
            a[i] = cls.gen_xi_op_xi_short(u_box + i, u_exp) & 0xffff;
        return a[:length]

    @staticmethod
    def invert_table(table, n_columns, len_result):
        assert len(table) & 31 == 0 and len_result & 31 == 0
        result = numpy.zeros(len_result, dtype = numpy.uint16) 
        for i, r in enumerate(table):
            if ((i & 31) < n_columns and (r & 0x7fff) < len_result):
                result[r & 0x7fff] = i | (r & 0x8000)
        return result


    @staticmethod
    def split_table(table, modulus):
        length = len(table)
        assert length & 31 == 0
        a = numpy.zeros(length >> 5, dtype = numpy.uint32) 
        for i in range(0, length, 32):
            sign = sum(((table[i+j] >> 15) & 1) << j for j in range(32))
            a[i >> 5] = sign
        table = (table & 0x7fff) % modulus
        return table, a




from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from math import log
from random import randint #, shuffle, sample
from functools import reduce
from operator import __or__

import numpy as np
import pytest

from mmgroup.structures.qs_matrix import QStateMatrix
from mmgroup.structures.qs_matrix import qs_unit_matrix, qs_rand_matrix
from mmgroup.tests.test_clifford.test_qs_product import create_m 
from mmgroup.tests.test_clifford.test_qs_product import rand_mul_scalar 

#####################################################################
# Create test matrices
#####################################################################




qs_matrix_data = [
    [ 0, 0, (1,0), []],
    [ 0, 4, (0,0), [0b00_0101, 0b00_0011]  ],
]



def create_testmatrices():
    """yield instances of class ``QStateMatrix`` for tests """
    for m_data in qs_matrix_data:
        m = create_m(*m_data)
        yield m
    yield qs_unit_matrix(4)  
    for i in range(20):
        yield qs_rand_matrix(2, 2, 4)
    for cols in range(0,7):
        for rows in range(0, 7):
            if rows + cols > 9:
                 break
            yield QStateMatrix(rows, cols)
            for data_rows in range(1, rows + cols + 3):
                for n in range(2):
                    m = qs_rand_matrix(rows, cols, data_rows)
                    rand_mul_scalar(m)  
                    yield m   
 

def bitrange_mask(imin, imax):
    """Return bit mask with bits set in ``range(imin, imax)``"""
    return (1 << imax) - (1 << imin)

def is_leading_low(x, j):
    """Return True if bit ``j`` is the lowest bit set in ``x``."""
    return x & ((1 << (j+1)) - 1) == 1 << j

def is_leading_high(x, j):
    """Return True if bit ``j`` is the lowest bit set in ``x``."""
    return x.bit_length() == j + 1


def zero_column(d, j, i):
    """Return True if ``d[k,j] = 0`` for ``0 <= k < i``

    Here ``d`` is a bit matrix implemented as an array of integers.
    """
    s = 0
    for k in range(i):
        s |= d[k]
    return (s & (1 << j)) == 0


#####################################################################
# Test method reduce_matrix()
#####################################################################


EPS = 1.0e-8

def show_row_table(text, m, row_tab):
    c = m.ncols
    t0, t1 = list(reversed(row_tab[:c])), row_tab[c:]
    s = "table of %s:" % text
    if len(t1):
        print(s, t0, t1[0], ", Q:", t1[1:])
    else:
        print(s, t0)


@pytest.mark.qstate
def test_reduce_matrix(verbose = 0):
    for ntest, m in enumerate(create_testmatrices()):
        # Check method m.reduce_matrix() for matrix m
        m1 = m.copy()
        if verbose:
            print("Test %d: reduce matrix, m =" % (ntest + 1), m)
            print("Standard reduced m =", m.copy().reduce())
        row_tab = m1.reduce_matrix()
        # Caution: 
        # The matrix-reduced representation of ``m1`` is very voltile,
        # since the user never actually needs it. Even ``__str__(m1)``
        # or ``m1 == _`` changes the representation of ``m1``
        # from 'reduced-matrix' to 'reduced'!!!
        m1.check()
        assert len(row_tab) == m1.nrows + m1.ncols
        if verbose:
            print("Reduced matrix, m1 =", m1)
            show_row_table("m1", m1, row_tab)
            print("shape", m1.shape)
        if  m != m1.copy():
            print("m reduced", m.copy().reduce())
            print("m1", m1.copy())
            print("m1 reduced", m1.display(reduced = False))
            err = "m and m1 are not equal"
            raise ValueError(err)
        if m1.nrows == 0: 
            assert  m.lb_rank() == -1
            continue
        fst_row = row_tab[m1.ncols]
        n0, n1 = m1.shape
        mask = (1 << m1.ncols) - 1
        mask += ((1 << m1.nrows) - 2) << m1.ncols
        d = [x & mask for x in m1.copy().check().raw_data[:m1.nrows]]
        imin = 0
        row_mask = bitrange_mask(0, n0 + n1)
        for j in range(n0 + n1 - 1, n1 - 1, -1):
            i =  row_tab[j]
            if i < 255: 
                assert i == imin + 1
                imin = i
                if not is_leading_high(d[i] & row_mask, j):
                    err = "error in row %d of m1"% i 
                    print("m1", m1.display(reduced = False))
                    show_row_table("m1", m1.copy(), row_tab)
                    raise ValueError(err)
        row_mask = bitrange_mask(0,n1)
        row_set = set(range(1, m1.nrows))
        for j in range(n1 - 1, -1, -1):
            i =  row_tab[j]
            ok = True
            if i < fst_row:
                ok = is_leading_high(d[i] & row_mask, j)
                #ok = ok and zero_column(d, j, i)
                assert row_tab[m1.ncols + i] == j
                row_set.remove(i)
            elif i < 255: 
                assert i == imin + 1
                imin = i
                ok = is_leading_high(d[i] & row_mask, j)
                #ok = ok and zero_column(d, j, fst_row)
                row_set.remove(i)
            if not ok:
                err = "error in column %d, row %d of m1" % (j, i) 
                print("m1", m1.display(reduced = False))
                show_row_table("m1", m1, row_tab)
                raise ValueError(err)
        assert imin == m1.nrows - 1
        for i in row_set:
            assert d[i] & row_mask == 0

        q_start, q_end = m1.ncols + fst_row, m1.ncols + m1.nrows
        q_mask = bitrange_mask(q_start, q_end)
        for j in range(q_start, q_end):
            i = row_tab[j]
            if i < 255:
                q_row = d[i] & q_mask
                assert i in row_set
                assert q_row & ~(1 << j) < 1 << j
                row_set.remove(i)
        for i in row_set:
            assert d[i] & q_mask == 0
            
        if m.ncols > 8:
            continue
        
        # Check the computation of the rank of matrix m              
        rk = np.linalg.matrix_rank(m[:], tol=EPS)
        lb_rk = m.lb_rank()
        assert rk == 1 << lb_rk, (m1[:], rk, lb_rk)   

        # Finally, check the computation of the norm of matrix m              
        n2 = np.linalg.norm(m[:], ord = 2)
        log_n2 = -1 if n2 == 0 else 2.0 * log(n2) / log(2.0)
        ok = abs(log_n2 - m.lb_norm2()) < EPS
        if not ok:
            print("Matrix m =", m.reduce())
            print("log_2(squared norm2) = %f, expected = %.13f" %
                (m.lb_norm2(), log_n2 ))
            err = "Error in computing norm of matrix"
            raise ValueError(err)            





import collections
import re
import warnings
from numbers import Integral
import numpy as np
from random import randint
from functools import partial

try:
    from mmgroup import mat24
    from mmgroup.mat24 import ploop_theta, pow_ploop, MAT24_ORDER
except:
    w = "Extension mmgroup.mat24 not found, package not functional!"
    warnings.warn(w, UserWarning)


from mmgroup.generators import rand_get_seed, gen_leech2_type
from mmgroup.generators import gen_rng_modp
from mmgroup.generators import gen_leech2_reduce_type4
from mmgroup.generators import mm_group_invert_word
from mmgroup.generators import mm_group_n_reduce_element
from mmgroup.generators import mm_group_n_clear
from mmgroup.generators import mm_group_n_mul_atom
from mmgroup.clifford12 import xsp2co1_rand_word_G_x0
from mmgroup.clifford12 import xsp2co1_rand_word_N_0

from mmgroup.structures.abstract_group import AbstractGroup
from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.parse_atoms import ihex, TaggedAtom, AtomDict
from mmgroup.structures.parse_atoms import  eval_atom_expression  
from mmgroup.structures.autpl import autpl_from_obj      
from mmgroup import Cocode, PLoop, AutPL



###########################################################################
# Standard atoms for the group M
###########################################################################


tag_dict = {
        "d": 0x10000000, 
        "p": 0x20000000, 
        "x": 0x30000000, 
        "y": 0x40000000, 
        "t": 0x50000000, 
        "l": 0x60000000, 
}




tags = " dpxytl"


ERR_ATOM_VALUE = "Illegal value of atom in constructor of monster group with tag '%s'"
ERR_ATOM_TYPE = "Illegal type '%s' of atom in constructor of monster group with tag '%s'"
ERR_TAG_VALUE = "Illegal tag '%s' in constructor of monster group"
ERR_TAG_TYPE = "Illegal type '%s' of tag in constructor of monster group"

def iter_d(tag, d):
    if isinstance(d, Integral ):
        yield 0x10000000 + (d & 0xfff)
    elif isinstance(d, str):
        cocode = randint(d == 'n', 0xfff) 
        if d == "o":
            cocode |= 1 
        elif d == "e":
            cocode &= ~1
        if  len(d) == 1 and d in "rnoe":
            yield 0x10000000 + (cocode & 0xfff)
        else:
            raise ValueError(ERR_ATOM_VALUE % 'd')            
    else:
        try:
            cocode = Cocode(d).cocode
            yield  0x10000000 + (cocode & 0xfff)
        except:
            raise TypeError(ERR_ATOM_TYPE % (type(d), 'd'))

def iter_p(tag, perm):
    if isinstance(perm, Integral):
        if not 0 <= perm < MAT24_ORDER:
            err = "Tag 'p': bad permutation number for Mathieu group M24"
            raise ValueError(err)
        yield 0x20000000 + perm
    elif isinstance(perm, str):
        perm_num = randint(perm == 'n', MAT24_ORDER-1)
        if  len(perm) == 1 and perm in "rn":
            yield 0x20000000 + perm_num
        else:
            raise ValueError(ERR_ATOM_VALUE % 'p')            
    elif isinstance(perm, AutPL):
        yield 0x10000000 + perm._cocode
        yield 0x20000000 + perm._perm_num
    else:
        try:
            cocode, perm_num = autpl_from_obj(0, perm)
            yield  0x10000000 + (cocode & 0xfff)
            yield  0x20000000 + perm_num
        except:
            raise TypeError(ERR_ATOM_TYPE % (type(perm), tag))


def gen_ploop_element(tag, r):
    if isinstance(r, Integral):
        return  r & 0x1fff
    elif isinstance(r, str):
        r_num = randint(r == 'n', 0x1fff)
        if  len(r) == 1 and r in "rn":
            return r_num
        raise ValueError(ERR_ATOM_VALUE % tag)            
    else:
        try:
            return PLoop(r).ord
        except:
            raise TypeError(ERR_ATOM_TYPE % (type(r), tag))

def iter_xy(tag, r):
    yield tag_dict[tag] + gen_ploop_element(tag, r)

def iter_z(tag, r ):
    pl = pow_ploop(gen_ploop_element(tag, r), 3)
    yield tag_dict['x'] + pl
    yield tag_dict['y'] + pl

def iter_tl(tag, r):
    if isinstance(r, Integral):
        e = r % 3
    elif isinstance(r, str):
        e = randint(int('n' in r), 2) 
        if  len(r) > 1 or not r in "rn":
            raise ValueError(ERR_ATOM_VALUE % tag)            
    else:
        raise TypeError(ERR_ATOM_TYPE % (type(r), tag))
    if e:
        yield  tag_dict[tag] + e 


###########################################################################
# Extended atoms for the group M
###########################################################################


ERR_Q_x0 = "Atom for tag 'q' must represent element of subgroup Q_x0 of the monster"


def iter_q(tag, r):
    if isinstance(r, Integral):
        e = r & 0x1ffffff
    elif isinstance(r, str):
        e = randint(r == 'n', 0x1ffffff) 
        if  len(r) != 1 or not d in "rn":
            raise ValueError(ERR_ATOM_VALUE % tag) 
    else:
        try:
            e = r.as_Q_x0_atom()
            assert isinstance(e, Integral)
        except:
            raise TypeError(ERR_Q_x0)
    d = (e >> 12) & 0x1fff
    delta = (e ^ ploop_theta(d)) & 0xfff
    yield tag_dict['x'] + d
    yield tag_dict['d'] + delta




ERR_LEECH2 = "Atom for tag 'c' must represent a type-4 vector in Leech lattice mod 2"


def rand_type4_vector():
    c = 0
    while gen_leech2_type(c) != 4:
        c = randint(0, 0xffffff) 
    return c

def iter_c(tag, r):
    if isinstance(r, str):
        if r == 'r':
            c = rand_type4_vector() 
        else:
            raise ValueError(ERR_LEECH2)
    elif isinstance(r, Integral):
        c = r & 0xffffff
    else:
        try:
            c = r.as_Q_x0_atom()
            assert isinstance(c, Integral)
        except:
            raise TypeError(ERR_LEECH2)
    if gen_leech2_type(c) != 4:
        raise ValueError(ERR_LEECH2)
    a = np.zeros(6, dtype = np.uint32)
    len_a = gen_leech2_reduce_type4(c, a)
    mm_group_invert_word(a, len_a)
    yield from a[:len_a]
     



###########################################################################
# Atoms generated from arrays of 32-bit integers
###########################################################################


ERR_TAG_A = "Atom for tag 'a' must be a list of unsigend 32-bit integers"

def iter_a(tag, a):
    try:
        for x in a:
            yield x & 0xffffffff
    except:
        raise TypeError(ERR_TAG_A)
       
###########################################################################
# Generating random atoms
###########################################################################


ERR_RAND_INTERN = "Internal error in generating random element of monster"

def iter_rand_N_0(in_N_x0, even):
    r"""Return random element of subgroup :math:`N_{0}`

    The function returns a uniform distributed random element
    of the subgroup :math:`N_{x}` of the monster of structure
    :math:`2^{2+11+22}.(M_{24} \times \mbox{Sym}_3)`. The group 
    :math:`N_0` is generated by the generators with tags
    ``x, y, d, p, t``. The function uses the internal random 
    generator of the ``mmgroup`` package.

    If parameter ``in_N_x0`` is nonzero then we compute a random
    element of the subgroup :math:`N_{x0}` of index 3 in :math:`N_0` 
    generated by the generators with tags ``x, y, d, p``.

    If parameter ``even`` is nonzero then we compute a random
    element of the  subgroup :math:`N_{\mbox{even}}` of index 2
    in :math:`N_{x}`  generated by the generators with
    tags ``x, y, d, p, t``, where all generators with tag ``d``
    correspond to even Golay cocode words.

    If both, ``in_N_x0`` and ``even``, are nonzero then we compute
    a random element
    of :math:`N_{xyz0} = N_{\mbox{even}} \cap N_{x0}`.
    """
    buf = np.zeros(10, dtype = np.uint32)
    seed = rand_get_seed()
    length = xsp2co1_rand_word_N_0(buf, in_N_x0, even, seed) 
    if not 0 <= length <= 10:
        raise ValueError(ERR_RAND_INTERN)
    yield from buf[:length]
     

def iter_rand_G_x0():
    r"""Return random element of subgroup :math:`G_{x0}`

    The function returns a uniform distributed random element
    of the subgroup :math:`G_{x0}`. The function uses the
    internal random generator of the ``mmgroup`` package.
    """
    buf = np.zeros(10, dtype = np.uint32)
    seed = rand_get_seed()
    length = xsp2co1_rand_word_G_x0(buf, seed) 
    if not 0 <= length <= 10:
        raise ValueError(ERR_RAND_INTERN)
    yield from buf[:length]



def iter_rand_mm(quality):
    r"""Return a random element of the monster group

    The function returns a random element of the monster group.
    Here ``quality`` means a measure for the quality of the
    randimization process, where a higher value menas that
    the distribution of the elements is closer to uniform.

    If ``quality`` is an integer ``k`` then a product containing
    ``k`` powers of the triality element it generated. Here the
    default value creates an almost uniform distribution.

    In future versions the default value of parameter ``quality``
    may correspond to the generation of a truly uniform
    distribution.
    """
    a = np.zeros(10, dtype = np.uint32)
    seed = rand_get_seed()
    len_a = xsp2co1_rand_word_G_x0(a, seed) 
    if not 0 <= len_a <= 10:
        raise ValueError(ERR_RAND_INTERN)
    yield from a[:len_a]
    for k in range(quality):
        yield 0x50000001 + gen_rng_modp(2, seed)
        c = 0
        while gen_leech2_type(c) != 4:
            c = gen_rng_modp(0x1000000, seed) 
        len_a = gen_leech2_reduce_type4(c, a)
        mm_group_invert_word(a, len_a)
        if not 0 <= len_a <= 6:
            raise ValueError(ERR_RAND_INTERN)
        yield from a[:len_a]



RAND_FUNCTIONS = {
    "M"      : (iter_rand_mm, 18),  
    "G_x0"   : (iter_rand_G_x0,),
    "N_0"    : (iter_rand_N_0, 0, 0), 
    "N_x0"   : (iter_rand_N_0, 1, 0), 
    "N_0_e"  : (iter_rand_N_0, 0, 1), 
    "N_x0_e" : (iter_rand_N_0, 1, 1), 
}

def iter_rand_subgroup(tag, s):
    if isinstance(s, Integral):
        yield from iter_rand_mm(s)
        return
    try:
        ff = RAND_FUNCTIONS[s]
    except KeyError:
        if isinstance(s, str):
            err = "Bad group name '%s' for tag 'r'"
            raise ValueError(err % s)
        else:
            err = "Atom for tag 'r' must be a string"
            raise TypeError(err)
    yield from ff[0](*(ff[1:]))
    
    

         

###########################################################################
# Generating atoms
###########################################################################


def iter_neutral(*args):
    return
    yield 0

gen_tag_dict = {
        "d": iter_d, 
        "p": iter_p, 
        "x": iter_xy, 
        "y" :iter_xy, 
        "z" :iter_z, 
        "t": iter_tl, 
        "l": iter_tl, 
        "q": iter_q, 
        "c": iter_c, 
        "a": iter_a, 
        "r": iter_rand_subgroup,
        "1": iter_neutral
}


def iter_atom(tag = None, number = None):
    """Return list of integers representing element of monster group

    This is the workhorse for method MMGroup.atom().
    See that method for documentation.
    """
    try: 
        gen_function = gen_tag_dict[tag]
    except KeyError:
        if isinstance(tag, str):
            raise ValueError(ERR_TAG_VALUE % tag)
        else:
             raise TypeError(ERR_TAG_TYPE % type(tag))
    yield from gen_function(tag, number)




###########################################################################
# Converting the input of a construtor of a monster element to atoms
###########################################################################


FRAME = re.compile(r"^([A-Za-z][A-Za-z_0-9]*)?\<(.+)\>$") 

EMBEDDED_GROUPS = [
]

EMBEDDED_CLASSES = (
   AutPL, Cocode,
)

TYPES_PARSED_FROM_LIST = EMBEDDED_CLASSES + (
   Integral, str, AbstractGroupWord,
)

ATOM_PARSERS = {
}

def add_to_embedded_classes(cls):
    global EMBEDDED_CLASSES
    if not cls in EMBEDDED_CLASSES:
        EMBEDDED_CLASSES = EMBEDDED_CLASSES + (cls,) 


def element_from_atom(group, tag, *data):
     if tag.isalpha():
         return group(tag, *data)
     else:
         raise ERR_TAG_VALUE(tag)


def iter_parse_mm_string(group, s):
    m = FRAME.match(s)
    group_name, string = (m[1], m[2]) if m else (None, s)
    if not group_name:
        group_name = group
    try:
        atom_dict = ATOM_PARSERS[group_name]
    except KeyError:
        if isinstance(group_name, str):
            err = "Cannot parse string of shape '%s<..>' in monster group" 
            raise ValueError(err % group_name)
        else:
            atom_dict = ATOM_PARSERS["M"] 
    element = eval_atom_expression(string, atom_dict)
    if element != 1:
        yield from element.mmdata



def load_group_name(group, group_name = None):
    global EMBEDDED_GROUPS, ATOM_PARSERS
    assert isinstance(group, AbstractGroup)
    EMBEDDED_GROUPS.append(group)
    atom_dict = AtomDict(partial(element_from_atom, group))
    ATOM_PARSERS[group] = atom_dict
    if group_name:
        assert isinstance(group_name, str) and group_name.isidentifier()
        ATOM_PARSERS[group_name] = atom_dict

import_groups_pending = True

def import_groups():
    global import_groups_pending
    global  MM, MMGroup, AbstractMMGroupWord
    from mmgroup import MM, MMGroup
    from mmgroup.structures.abstract_mm_group import AbstractMMGroupWord
    #load_group_name(MM().group, "M")  # TODO: Fixme!!!!
    import_groups_pending = False




def iter_mm(group, tag = None, atom = None, in_G_x0 = False):
    if isinstance(tag, str) and len(tag) == 1:
        yield from iter_atom(tag, atom)
        return
    if tag is None:
        return
    if import_groups_pending:
        import_groups()
    if isinstance(tag, AbstractMMGroupWord):
        yield from tag.mmdata
    elif isinstance(tag, EMBEDDED_CLASSES):
        yield from tag.mmdata
    elif isinstance(tag, str):
        if group is None:
            group = MMGroup
        yield from iter_parse_mm_string(group, tag)
    elif isinstance(tag, list):
        for element in tag:
            if isinstance(element, tuple):
                yield from iter_atom(*element)
            elif isinstance(element, TYPES_PARSED_FROM_LIST):
                yield from iter_mm(element)
            else:
                raise TypeError(ERR_TAG_TYPE % type(element))
    elif isinstance(tag, Integral):
        if tag == 1:
            return
        elif tag == -1 and in_G_x0:
            yield 0x31000000
            return
        err = "Cannot convert integer to monster group element"
        raise ValueError(err)
    else:
        raise TypeError(ERR_TAG_TYPE % type(tag))



def print_mm_atoms(tag = None, data = None, group = None):
    print([hex(x) for x in iter_mm(group, tag, data)])



###########################################################################
# Converting atoms to tuples
###########################################################################


ERR_ILLEGAL_ATOM = "Illegal atom %s in monster group element"

def iter_tuples_from_N_x0_element(element):
    mm_group_n_reduce_element(element)
    if element[1]: yield('y', element[1])
    if element[2]: yield('x', element[2])
    if element[3]: yield('d', element[3])
    if element[4]: yield('p', element[4])
    mm_group_n_clear(element)


def iter_tuples_from_atoms(atoms):
    N_x0_element = np.zeros(5, dtype = np.uint32)
    for a in atoms:
        a &= 0xffffffff
        t = a & 0x7fffffff
        if t < 0x50000000:
            mm_group_n_mul_atom(N_x0_element, a)
        elif a & 0xfffffff == 0:
            continue
        elif t < 0x70000000:
            tag, v = t >> 28, a & 0xfffffff
            v = (-v if a & 0x80000000 else v) % 3
            if v:
                yield from iter_tuples_from_N_x0_element(N_x0_element)
                yield ("tl"[tag-5], v)
        else:
            raise ValueError(ERR_ILLEGAL_ATOM % hex(a))
    yield from iter_tuples_from_N_x0_element(N_x0_element)


def print_mm_tuples(tag = None, data = None, group = None):
    tuples = iter_tuples_from_atoms(iter_mm(group, tag, data))
    print([hex(x) for x in tuples])


###########################################################################
# Converting atoms to strings
###########################################################################


ERR_ILLEGAL_ATOM = "Illegal atom %s in monster group element"

def iter_strings_from_N_x0_element(element):
    mm_group_n_reduce_element(element)
    if element[1]: yield "y_%s" % ihex(element[1])
    if element[2]: yield "x_%s" % ihex(element[2])
    if element[3]: yield "d_%s" % ihex(element[3])
    if element[4]: yield "p_%d" % element[4]
    mm_group_n_clear(element)


def iter_strings_from_atoms(atoms, abort_if_error = True):
    N_x0_element = np.zeros(5, dtype = np.uint32)
    for a in atoms:
        a &= 0xffffffff
        t = a & 0x7fffffff
        if t < 0x50000000:
            mm_group_n_mul_atom(N_x0_element, a)
        elif a & 0xfffffff == 0:
            continue
        elif t < 0x70000000:
            tag, v = t >> 28, a & 0xfffffff
            v = (-v if a & 0x80000000 else v) % 3
            if v:
                yield from iter_strings_from_N_x0_element(N_x0_element)
                yield "%s_%d" % ("tl"[tag-5], v)
        else:
            if abort_if_error:
                raise ValueError(ERR_ILLEGAL_ATOM % hex(a))
            yield "<Bad atom %s>" % hex(a)
    yield from iter_strings_from_N_x0_element(N_x0_element)


def print_mm_string(tag = None, data = None, group = None):
    strings = iter_strings_from_atoms(iter_mm(group, tag, data))
    s = "*".join(strings)
    print(s if len(s) else "1")



###########################################################################
# Converting tuples to strings
###########################################################################


TUPLE_FMT_DICT = dict(zip("dpxytl", [ihex, str, ihex, ihex, str, str]))

def iter_strings_from_tuples(tuples):
     for tag, value in tuples:
          try:
              fmt = TUPLE_FMT_DICT[tag]
          except KeyError:
              if isinstance(tag, str):
                  err = "Illagal tag in tuple of monster group atoms"
                  raise ValueError(err)
              else:
                  err = "Tag in tuple of monster group atoms must be a string"
                  raise TypeError(err)
          yield "%s_%s" % (tag, fmt(value))




r"""The module contains classes ``Lsbit24Function`` and ``MatrixToPerm``.

Class ``Lsbit24Function`` contains tables and directives for computing
the least significant bit of a ``24-bit`` integer using a
DeBruijn sequence. This may be overkill for such a simple function,
but it may also be considered as a didactic example of a 
table-providing class.

Class ``MatrixToPerm`` contains a directive that generates highly 
optimized code for converting a bit matrix acting on a Golay code 
word to a permutation in the Mathieu group ``Mat_24``. Here we 
assume the bit matrix actually encodes an element of the  
Mathieu group; otherwise garbage is returned. The Golay code
word must be given as an integer in ``gcode`` representation.
"""
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

# python 3 string formatting see:
# https://pyformat.info/


import types
import sys
import re
import os
from random import randint
from operator import __or__, __xor__, __and__
from functools import reduce

import numpy
from numpy import array, zeros, uint8, uint16, uint32


from mmgroup.bitfunctions import bw24, bits2list, iter_bitweight, v2
from mmgroup.bitfunctions import iter_ascending_bitweight, lmap



from mmgroup.generate_c import prepend_blanks, format_item
from mmgroup.generate_c import UserDirective, UserFormat



   


 


############################################################################
############################################################################
# Generating C code for bit functions
############################################################################
############################################################################


def lsbit24(x):
    """Return position of least significant bit of (x & 2**24)"""
    return v2(x | 0x1000000)


def _make_de_bruijn_sequence():
    """Create DeBruijn sequence for finding the lowest bit of an integer

    This is a modification of
    http://graphics.stanford.edu/~seander/bithacks.html#ZerosOnRightMultLookup
    which returns 24 if x & 0xffffff is zero.

    Here we just inspect the lower 24 bit of an integer and return 24
    if they are all zero.
    """
    lsbit_table = numpy.zeros(32, dtype = uint8)
    sequence = 0x077CB531
    for i in range(32):
        index =  ((sequence << i) >> 26) & 0x1f
        lsbit_table[index] = min(i,24)  
    assert lsbit_table[0] == 24
    return sequence,  lsbit_table


class Lsbit24Function(object):
    """Generate code for computing function ``lsbit24()``

    The class uses the ideas in function ``_make_de_bruijn_sequence()``
    for computing that  least significant bit of an integer.

    This documentation is intended to explain the cooperation of this
    class with the class ``TableGenerator`` in module
    ``mmgroup.generate_c``. Classes designed 
    for a similar cooperation may refer to this documentation.

    The class requires a table ``lsbit_table``, which is generated 
    by function ``_make_de_bruijn_sequence()``.

    This is clearly to much overhead fur such a tiny function, be we
    use this as an example for documenting an application of the code-
    generating class ``TableGenerator``.


    This class hat the following public member functions

    ================  ==================================================
    ``tables()``      Returns a dictionary containing the tables
                      required for generating the C code
    ``directives()``  Returns a dictionary containing directives 
                      required for generating the C code    
    ================  ==================================================

    This class is designed for cooperation with an instance of
    class ``TableGenerator``. A similar class should provode at least 
    the member functions ``tables()`` and ``directives()``.

 
    In order to create the table ``lsbit_table``, put e.g. the lines::

        // %%USE_TABLE
        static const uint8_t MAT24_LSBIT_TABLE[32] = { 
        // %%TABLE Mat24_lsbit_table, uint8
        };

    into the source file. This creates the table ``MAT24_LSBIT_TABLE``
    for  computing the lsbit quickly. 
 
    This class provides a function with name ``MAT24_LSBIT`` for 
    generating  C code that compute the lsbit of ``x``. After generating 
    the table given above, the expression ``{MAT24_LSBIT:x}`` in the input 
    file generates a very efficient expression that evaluates to the 
    lsbit of ``x``. 
    Note that we use standard python formatting syntax, with 
    ``MAT24_LSBIT`` refering to a member  function ``generate`` of class 
    ``Lsbit24Function``that generates the appropriate C code for 
    computing the lsbit of the 24-bit integer ``x``.
 
    The entry 

          "MAT24_LSBIT" : UserFormat(cls.generate, "s", prefix=True)
    
    in the attribute ``tables`` means the following:

    We create an instance of class ``mmgroup.generate_c.UserFormat``
    that behaves like member function ``generate`` of this class.
    Furthermore, this instance has a ``.format`` method so that
    the string ``'{MAT24_LSBIT:x}'`` in the source file for code 
    generation  evaluates to the string returned by applying member 
    function  ``Lsbit24Function.generate`` to the argument ``'x'``.
    Here function ``Lsbit24Function.generate`` returns a string with
    very efficient C code for computing the lsbit of variable ``x``.
    
    Note that python and C name spaces are strictly separated. So the
    table coded above has the name  ``Mat24_lsbit_table`` in python.
    In other words, the python object ``Mat24_lsbit_table`` contains the 
    data that make up the table. In the generated C file, the table 
    may have another name (which is ``MAT24_LSBIT_TABLE`` in the example
    above). The instance of the code generation class provides a 
    dictionary ``names`` that maps python names to C names. Dictionary 
    ``names`` must be passed to the member function ``generate`` as the 
    first parameter, so the member function ``generate`` can 
    infer the C name of the table from its python name 
    ``Mat24_lsbit_table``. The keyword argument ``prefix=True`` makes
    sure that dictionary ``names`` is prepended to the list of arguments
    for function ``Lsbit24Function.generate`` given by the user.

    The second argument ``"s"`` of ``UserFormat()`` means that the
    argument (which is ``x`` in the example ``{MAT24_LSBIT:x}``) is
    interpreted as the string ``"x"". See documentation of class 
    ``UserFormat()`` for details.

    """
    CONST,  table = _make_de_bruijn_sequence()
    hCONST = format_item(CONST, "uint32")
   
    @classmethod
    def compute(cls, x):
        """Return index of least significant bit of an integer x

        If x & 0xffffff is zero then 24 is returned.        
        """
        x = int(x & 0xffffffff)
        return int(cls.table[(((x & -x) * cls.CONST) >> 26) & 0x1f])



    @classmethod
    def generate(cls, names, x, arg_is_pwr2 = False):
        """generate C code computing  lsbit24(x)

        Here 'x' is a string representing the input as C code.

        When calling this function from python, names is the
        dictionary for translating python names of tables into
        C names to be used by the generated code, see documention
        of this class.

        If 'arg_is_pwr2' then more efficient code is generated,
        assuming that x is a power of two.

        """
        if not arg_is_pwr2:
             x = "({0}) & (0-({0}))".format(x)
        return  """{1}[({2} *  \\
        ({0}) >> 26) & 0x1f]""".format(x, 
            names["Mat24_lsbit_table"], cls.hCONST)


    @classmethod
    def tables(cls):
        """Return dictionary of tables for class TableGenerator

        """ 
        return { 
            "Mat24_lsbit_table" :  cls.table[:], 
            "MAT24_LSBIT" : UserFormat(cls.generate, "s", prefix=True),
        }

    @classmethod
    def directives(cls):
        """No directives here""" 
        return {} 
        


############################################################################
############################################################################
# Functions for analyzing basis vecors of the Golay code  
############################################################################
############################################################################
 
def split_golay_codevector(v, check=None):
    """Split Golay code vector into 'blackwhite' and 'colored' part
 
    The 'blackwhite' part has the same value (0 or 1) in the bottom
    three entries of each MOG column. 

    The 'colored' part has 0 or 2 entries with value 1 in the bottom
    three entries of each MOG column, and no entry with value 1 in
    the top row. 

    Returns pair of Golay code vectors ('blackwhite', 'colored ')
    with the sum of these vectors equal to v.

    If 'check' is set, it must behave lie an instance of class GolayCode.
    """
    color = [0, 6, 5, 3, 3, 5, 6, 0] 
    col = 0
    for i in range (1, 25, 4):
         col |= color[(v >> i) & 7] << i
    v ^= col
    if check:
        cn = check.vect_to_vintern(v) | check.vect_to_vintern(col) 
        assert cn & 0xfff == 0, "not a Golay code word"
    return v, col





############################################################################
############################################################################
# Functions for changing Colay code rep to permutation rep in Mat24
############################################################################
############################################################################


def generate_golay24_decode(table_names, s): 
        """Generate code for converting a Golay code word 's'.

        Generate C code for computing the 'vect' representation of a
        Golay code word 's' given in 'gcode' representation. See class
        Golay24.GolayCode for background.

        The generated code is and expression containing the input 's'
        and conversion tables from class Golay24.GolayCode.
        """ 
        return """({1}[({0} << 4) & 0xf0]
       ^ {2}[({0}  >> 4) & 0xff])
""".format(s, table_names["Mat24_dec_table1"],
                table_names["Mat24_dec_table2"])



class MatrixToPerm(object):
    """Support for conversion of Mat24 from matrix to permutation rep

    A basis of the Golay code is given with the constructor.
    The basis should contain 6 colored and 6 blackwhite basis vectors,
    exactly one of the blackwhite basis vectors should be odd.

    The basis is augmented by up to two colored vectors of weight 12,
    such that their XOR sum has also weight 12. Call these vectors
    c0 and c1. Then we calculate c0 & c1, c0 & ~c1, ~c0 & c1. These 3
    vectors, and the odd blackwhite vector and its complement form a 
    set of 5 vectors. Any singelton vector in the MOG an be obtained 
    as an intersection of one of these 5 vectors with a vector that 
    has ones in exactly one column of the MOG.

    Then the even blackwhite basis vectors are processed with 
    operators 'and' and 'not' so that for each MOG column there is a
    vector t[i]  that contains ones in precisely one column. 

    If an image of the standard basis of the Golay code is given,
    we can perform these XOR, AND and coplement operations with the 
    transformed basis vectores, yielding the transformed singletons.

    These transformed singletons are just the images of an element
    of the Mathieu group Mat24 considered ad a permutation. 

    Member functions tables(), directives() and bytelen() and also
    compute() and generate() follow the conventions in 
    class Lsbit24Function. 
    """
    def __init__(self, basis, verbose = 0):
        """Set basis for the Golay code and calculate tables
      
        separate (even and odd) blackwhite and colored basis vectors
        """ 
        self.basis = list( basis )
        self.verbose = verbose
        self.split_basis()   
        self.augment_colored_basis() 
        self.make_aux_table()
        self.test(verbose=0)
        #print (self.generate("b", "p"))


    def split_basis(self):
        """Split basis into blackwhite and colored vectors.

        self.blackwhite =  list of indices of even blackwhite basis vectors
        self.colored =     list of indices of colored basis vectors
        self.odd_v =       index of the (single) odd blackwhite basis vector
        """
        self.blackwhite = []
        self.colored = []
        self.odd_v = None
        for i, v in enumerate(self.basis):
            bw, col =  split_golay_codevector(v)
            if v == col:
                self.colored.append(i)
            if v == bw:
                if bw24(v & 0x111111) & 1:
                    self.odd_v = i
                else:
                    self.blackwhite.append(i)


    def and_basis(self, z):
        """Compute and operation (with complements) on basis vectors

        z is a list of pairs (index, sign). self.basis[index] is complemented
        if and only if sign is 1, and all these (possibly complemented)
        basis vectors are combined with the and operation.
        The result is returned. 
        """
        return  reduce(__and__, [self.basis[pos] ^ -sign for (pos,sign) in z])
       

    def augment_colored_basis(self):
        """Augment self.basis by addding some vectors to the basis.
        
        The augmented vectors are XOR combinations of the basis vectors.
       
        The following tables are computed:

        self.basis_augment is a list of entries, where each entry represents
        an augmented basis vector as a list of the indices of basis vectors.
        The corresponding basis vectors are to be XORed to obtain the new
        basis vector.    

        self.col_aux_table is a list of 5 entries. Here entry i is the
        instruction how to compute the i-th auxiliary vector aux[i] from 
        the basis vectors with member function  and_basis.

        self.out_table is a table of 24 auxiliary vectors. For 
        obtaining the i-th singleteon we have to compute

            col[i >> 2] &  aux[self.out_table[i]] .
        
        Here col[i] is the vector that casts out the entries in the
        i-th column in the MOG. A table for computing col[i] is prepared
        in member function make_aux_table.
        """
        colored = self.colored
        basis_list = []
        value_list = []
        colored_6 = []
        self.basis_augment = []
        done = False
        for pos in lmap( bits2list,iter_ascending_bitweight(1 << len(colored))):
            basispos = [ colored[i] for i in pos]
            value = reduce(__xor__, [self.basis[p] for p in basispos], 0)
            if bw24(value) == 12:
                 basis_list.append(basispos)
                 value_list.append(value)
                 for i, val in enumerate(value_list):
                     if bw24(value ^ val) == 12:
                         for j in [i, -1]:
                              b,v = basis_list[j], value_list[j]
                              if len(b) > 1:
                                  self.basis_augment.append(b)
                                  b = [len(self.basis)] 
                                  self.basis.append(v)
                                  self.colored.append(b)
                              colored_6.append(b[0])
                         done = True
                         break
            if done: break

        c0, c1 = colored_6[0], colored_6[1]
        self.col_aux_table = [
             [(self.odd_v,0)],  [(self.odd_v,1)],
             [(c0,0), (c1,0)], [(c0,0), (c1,1)], [(c0,1), (c1,0)]
        ]
        col_aux_values = lmap(self.and_basis, self.col_aux_table)
        self.out_table = []
        for i in range(24):
            mask = 0xf << (i & -4)
            for j, v in enumerate(col_aux_values):
                 if v & mask == 1 << i:
                     self.out_table.append(j)
                     break
        assert len(self.out_table) == 24
        if self.verbose:
            print("augmented basis for Golay code to perm rep of Mat24") 
            print(lmap(hex, self.basis))
            print("augment:", self.basis_augment, "col", colored_6)
            print("col_aux_table:", self.col_aux_table)
            print("values:", [hex(x & 0xffffff) for x in col_aux_values])
            print("t_out:", self.out_table)



    def make_aux_table(self):
        """Prepare table for casting out the i-th column in the MOG

        Let col[i], i=0,...,5 is the vector that casts out the entries 
        in the i-th column in the MOG. col[i] is computed from the 
        basis vectors in self.basis with function and_basis().
         
        self.bw_aux_table is a table with 6 entries, where entry i
        is used to compute  col[i] by member function make_aux_table.
        """
        blackwhite = self.blackwhite
        lbw = len(blackwhite) 
        bw_combinations = [None] * 6
        bw_dict = dict(zip([0xf << (4*i) for i in range(6)], range(6))) 
        done = False
        for k in range(1,lbw+1):
            for sgn in iter_ascending_bitweight((1 << k)-1):
                signs = [ (sgn >> i) & 1 for i in range(k)]
                for pos in lmap(bits2list,iter_bitweight(k,1 << lbw)):
                    basispos = [ blackwhite[i] for i in pos]
                    z = list(zip(basispos,signs))
                    res = self.and_basis(z)
                    try:
                        bw_combinations[bw_dict[res]] = z
                        del  bw_dict[res]
                        done = not None in bw_combinations
                        if done: break
                    except:
                        pass
                if done: break
            if done: break
        self.bw_aux_table = bw_combinations 
        if self.verbose:
            print("bw_comb",  bw_combinations ) 


    def compute(self, basis):
        """Calculate the permutation from a (transformed) basis.

        Here we calculate the singletons from the transformed basis.
        The same calculations with the standard basis self.basis yields 
        the singletons  0,...,23 in that order. So the singletons from 
        the transformed basis are the images of 0,...,23. 

        The tables self.basis_augment, self.col_aux_table, 
        self.bw_aux_table, self.out_table[i] are used.

        Singleton i is represented as 2**i, so at the end we have to take
        binary logarithms.
        """
        basis = basis[:12]
        for l in self.basis_augment:
            basis.append(reduce(__xor__, [basis[x] for x in l], 0))
        t = []
        for z in self.col_aux_table + self.bw_aux_table:
            t.append(reduce(__and__, [basis[i] ^ -s for (i,s) in z]))
        out = []
        st = len(self.col_aux_table)
        for i in range(24):
            m =  t[st + (i>>2)] & t[self.out_table[i]]
            out.append( v2(m) )
        return out

    def test(self, verbose = 0):
        """A simple self test. 

        self.compute(self.basis) must return [0,...,23].
        """
        basis = self.basis[:12]
        result = self.compute(basis)
        assert result == list(range(24)), result 
        if verbose:
            print("Test of class Matrix_to_perm passed." )


    def generate_and(self, z, basis_name):
        """Return C code for the expession self.and_basis(z)

        basis_name is the name of the array for the basis         
        """
        v = [[], []]
        for i,s in z:
            v[s].append("%s[%d]"  % (basis_name, i))
        neg = " | ".join(v[1])
        if len(neg): v[0].append("~(%s)" % neg)
        return " & ".join(v[0])

    def generate(self, names, in_name, out_name, 
             basis_name = "ba",  temp_name = "t"):
        """Generate C code roughly corresponding to self.calculate()

        'in_name' is the name of the input array representing the transformed
        basis. Each entry of that input array represent a Golay code word.
        In contrast to function self.calculate(), the Golay code vectors in
        the input array are given in 'gcode' represention, see documentation
        of class GolayCode. They are converted to vector representation first.
        
        'out_name' is the name of the output array representing the 
        resulting element of the Mathieu group Mat24 as a permutation.

        The remaining parameters are optional names for some internal arrays.
        """
        self.names = names
        s =  """uint_fast32_t _i, {0}[14], {1}[{2}];
for (_i=0; _i < 12; ++_i) {0}[_i] =
    {3};
""".format(
           basis_name, temp_name, 
           len(self.col_aux_table + self.bw_aux_table),
           generate_golay24_decode(names, "%s[_i]" % in_name)
        )


        for i,l in enumerate(self.basis_augment):
            v = " ^ ".join( ["%s[%d]" % (basis_name, j)  for j in l] )
            s += "%s[%d] = %s;\n" % (basis_name, i+12, v)
        for i, z in enumerate(self.col_aux_table + self.bw_aux_table):
            v = self.generate_and(z, basis_name)
            s += "%s[%d] = %s;\n" % (temp_name, i, v)

        t = [sum( x << (4*j) for j, x in enumerate(self.out_table[i:24:4]) )
                  for i in range(4)  ]
        s1 = "uint_fast32_t _w, _k = %s[_i + %d];\n" % (
                               temp_name, len(self.col_aux_table))
        for j in range(4):
            s1 += "_w =  %s[(%s >> (_i << 2)) & 0xf] & _k;\n" % (
                  temp_name, hex(t[j]) ) 
            # Next put *<out_name>++ = lsbit24(_w)
            lsb =  Lsbit24Function.generate(names, "_w", arg_is_pwr2=True)
            s1 +=  "*%s++ = %s;\n" % (out_name, lsb);

        s += "for (_i = 0; _i < 6; ++_i) {\n"
        s += prepend_blanks(s1,3)
        s += "}\n"
        s += "%s -= 24;\n" % out_name
        return s  
       
    def tables(self):
        """This class generates no tables"""
        return {}

    def directives(self):
        """Code generation function has name "MAT24_MATRIX_TO_PERM". """
        return   {
            "MAT24_MATRIX_TO_PERM" : UserDirective(self.generate,"ssss", 1),
        }
 




r"""We deal with octads and certain subsets of octads called suboctads.

An *octad* is a  Golay code word of weight ``8``. The Golay code 
has ``759`` octads. A signed octad is a Parker loop element 
corresponding to an octad or a complement of an octad. Signed 
octads are used for indexing some unit vectors of the 
representation :math:`\rho` of the monster :math:`\mathbb{M}`.

Function ``Octad()`` returns a signed octad as an instance of 
class |PLoop|. Arguments of function ``Octad()`` are interpreted
in the same way as arguments of the constructor for class |PLoop|,
but the function raises `ValueError` if the result is not a
(possibly complemented) signed octad.

We use some lexical order for numbering the ``759`` octads,
which we do not describe in detail.  
Due to the well-known properties of the Golay code we can create 
a random octad and display its number as follows:

.. code-block:: python
    
        from random import sample
        from mmgroup import Octad
        # Select 5 random integers between 0 and 23
        int_list = sample(range(24), 5)
        # Complete these 5 integers to a (unique) octad o
        o = Octad(int_list)
        # Display octad o and its number
        print("Octad", o.bit_list, "has number", o.octad)


A *suboctad* of an octad ``o`` is an element of the Golay cocode
:math:`\mathcal{C}^*` of even parity which can be represented
as a subset of ``o``. Each octad has ``64`` suboctads.
A pair (octad, suboctad) is used for indexing some basis
vectors in the representation :math:`\rho`. function ``SubOctad``
creates a suboctad as an instance of class |XLeech2| from such a 
pair. Function ``SubOctad`` takes two arguments ``octad`` and 
``suboctad``. The first argument evaluates to a signed octad as 
in function ``Octad()``. The second argument evaluates to a 
suboctad, see  function ``SubOctad`` for details.

The raison d'etre of a  suboctad is indexing a basis vector in
the representation  :math:`\rho`. For this purpose we need a pair 
of integers refering to the octad and the suboctad. For an instance 
``so`` of class |XLeech2| that pair is given as the pair of the
last two integers in ``so.vector_tuple()``.

For an octad ``o`` and a suboctad ``s`` precisely one of the
pairs ``(o,s)`` and ``(~o,s)`` is actually used as an index for 
:math:`\rho`. The pair ``(~o,s)`` is used if the cocode element 
corresponding to ``s`` has minimum weight ``2``; the pair 
``(o,s)`` is used if that element has minimum weight ``0`` or 
``4``. See :cite:`Con85` or :cite:`Seysen20` for background. In 
this  implementation both, ``(o,s)`` and ``(~o,s)``, refer to the 
same basis vector of :math:`\rho`.
"""



from functools import reduce
from operator import __xor__
from numbers import Integral, Number
from random import randint
import warnings


try:
    # Try importing the fast C function
    from mmgroup import mat24 
except (ImportError, ModuleNotFoundError):
    # Use the slow python function if the C function is not available
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24



from mmgroup.structures.abstract_group import AbstractGroupWord
from mmgroup.structures.parity import Parity
from mmgroup.structures.parse_atoms import ihex

from mmgroup.structures.autpl import AutPL, AutPlGroup

from mmgroup.structures.gcode import GCode, GcVector
from mmgroup.structures.cocode import Cocode
from mmgroup.structures.ploop import PLoop


ERR_RAND = "Illegal string for constricting type %s element" 

ERR_DIV4 = "%s object may be divided by 2 or 4 only"
ERR_DIV2 = "%s object may be divided by 2 only"



#######################################################################
# Import derived classed
#######################################################################

import_pending = True

def complete_import():
    """Internal function of this module

    If you need any of the objects declared above, do the following:

    if import_pending:
        complete_import()
    """
    global import_pending,  XLeech2
    from mmgroup.structures.xleech2 import XLeech2
    import_pending = False



#######################################################################
# Auxiliary functions
#######################################################################



def as_vector24(data):
    """Convert list of bit positions to vector in ``GF(2)**24``"""
    try:
        vector = reduce(__xor__, (1 << x for x in data), 0)
    except:
        err = "List of integers >= 0 required for a 24-bit vector"
        raise TypeError(err)
    if vector >= 0x1000000:
        raise ValueError("List entry too large for a 24-bit vector")
    return vector


def str_vector24(v):
    """Return the 24-bit vector ``v`` as a string of binary digits

    Here ``v`` must be an integer with the coefficient of the
    ``i``-th bit coded in the bit with valence ``2**i``.
    Coefficients  ``i >= 24`` are ignored.
    """
    l = ["01"[(v >> i) & 1] for i in range(24)]
    l4 = ("".join(l[i:i+4]) for i in range(0, 24, 4))
    return " ".join(l4)

def str_basis(text, letter, basis):
   s = text + "\n"
   for i, v in enumerate(basis):
       s += "  %3s: %s\n" % (letter + str(i), str_vector24(v)) 
   return s


#######################################################################
# Function Octad
#######################################################################







def Octad(octad):
    """Return a (signed) octad as an element of the Parker loop. 

    :param octad:  the value of the octad to be returned. 
    :type octad: see table below for legal types

    :return:  a (signed) octad
    :rtype:   an instance of class |PLoop|

    :raise:
        * TypeError if ``type(octad)`` is not in the table given above.
        * ValueError if ``octad`` cannot be converted to a
          (possibly negated and complemented) octad.


    Depending on its type parameter **octad** is  interpreted as follows:

    .. table:: Legal types for parameter ``octad``
      :widths: 20 80

      ===================== ================================================
      type                  Evaluates to
      ===================== ================================================
      ``int``               Here the (positive) octad with  number 
                            ``octad`` is returned. There are 759 octads
                            in the Golay code. 
                            So ``0 <= octad < 759`` must hold.
                            

      ``list`` of ``int``   Such a list is converted to a Golay code word,
                            see class |GCode|, and the corresponding 
                            (positive) Parker loop element is returned.

      class |GCode|         The corresponding 
                            (positive) Parker loop element is returned. 

      class |PLoop|         A deep copy of parameter ``octad`` is returned.

      class |GcVector|      This is converted to a Golay code word,
                            see class |GCode|, and the corresponding 
                            (positive) Parker loop element is returned.

      class |XLeech2|       The *octad* part of the vector ``octad``  
                            is  returned. 

      ``str``               Create random element depending on the string
                             | ``'r'``: Create arbitrary octad

      ===================== ================================================

    A complement of an octad is also accepted; then the corresponding 
    Parker loop element is retured. The function raises ValueError
    if  parameter ``octad`` does not evaluate to an octad or a
    complement of an octad.      
    """
    if isinstance(octad, Integral):
        if  not 0 <= octad < 759:
            raise ValueError("Bad octad number")
        return PLoop(mat24.octad_to_gcode(octad) & 0xfff)
    elif isinstance(octad, str):
        return PLoop(mat24.octad_to_gcode(randint(0, 758)))
    elif isinstance(octad, PLoop):
        _ = octad.octad
        return octad
    elif isinstance(octad, GCode):
        _ = octad.octad
        return PLoop(octad.ord)
    else:
        v =  PLoop(octad)
        _ = v.octad
        return v





#######################################################################
# Function  SubOctad
#######################################################################


def SubOctad(octad, suboctad = 0):
    """Creates a suboctad as instance of class |XLeech2|

    :param octad:

      The first component of the pair *(octad, suboctad)* to be 
      created.

    :type octad: same as in function
      :py:class:`~mmgroup.Octad`

    :param suboctad:
    
      The second component of the pair 
      *(octad, suboctad)* to be created.

    :type suboctad: see table below for legal types


    :return: 

     The suboctad given by the pair *(octad, suboctad)* 

    :rtype: an instance of class |XLeech2|

    :raise:
        * TypeError if one of the arguments *octad* or *suboctad* 
          is not of correct type.
        * ValueError if argument *octad* or *suboctad* does not
          evaluate to an octad or to a correct suboctad,
          respectively.
    

    A *suboctad* is an even cocode element that can be represented
    as a subset of the octad given by the argument *octad*. 

    The raison d'etre of function ``SubOctad`` is that pairs
    *(octad, suboctad)* are used for indexing vectors in the
    representation of the monster group. Here we want to 
    number the octads from ``0`` to ``758`` and the suboctads
    form ``0`` to ``63``, depending on the octad. Note that
    every octad has ``64``  suboctads.

    Depending on its type parameter **suboctad** is  interpreted as follows:

    .. table:: Legal types for parameter ``suboctad``
      :widths: 20 80

      ===================== ================================================
      type                  Evaluates to
      ===================== ================================================
      ``int``               Here the suboctad with the number given 
                            in the argument is 
                            taken.  That numbering depends on the octad 
                            given in   the argument ``octad``. 
                            ``0 <= suboctad < 64`` must hold.                           

      ``list`` of ``int``   Such a list is converted to a bit vector
                            as in class |GcVector|,
                            and the cocode element corresponding to that
                            bit vector is taken.

       class |GCode|        The intersection of the octad given as the 
                            first argument and the Golay code word given
                            as the second argument is taken. 
  
       class |GcVector|     This is converted to a cocode element,
                            see class |Cocode|, and that cocode element 
                            is taken.

       class |Cocode|       That cocode element is taken as the suboctad.


      ``str``               Create random element depending on the string
                             | ``'r'``: Create arbitrary suboctad
      ===================== ================================================

    The numbering of the suboctads

    Suboctads are numbered for 0 to 63. Let ``[b0, b1,..., b7]`` be the 
    bits set in the octad of the pair ``(octad, suboctad)`` in natural 
    order. The following table shows the suboctad numbers for some 
    suboctads given as cocode elements. More suboctad numbers can be 
    obtained by combining suboctads and their corresponding numbers 
    with ``XOR``.

    .. table:: Suboctad numbers of some cocode elements
      :widths: 16 16 16 16 18 18

      =========== =========== =========== =========== =========== =========== 
      ``[b0,b1]`` ``[b0,b2]`` ``[b0,b3]`` ``[b0,b4]`` ``[b0,b5]`` ``[b0,b6]``
      ``s  = 1``  ``s  = 2``  ``s  = 4``  ``s  = 8``  ``s = 16``  ``s = 32``
      =========== =========== =========== =========== =========== =========== 

    E.g. ``[b0, b5, b6, b7]`` is equivalent to ``[b1, b2, b3, b4]`` 
    modulo the Golay code and has number ``s = 1 ^ 2 ^ 4 ^ 8 = 15``.

    """
    ploop = Octad(octad)
    gcode = ploop.value & 0xfff
    if isinstance(suboctad, str):
        suboctad_ = randint(0, 63)
    elif isinstance(suboctad, Integral):
        suboctad_ = suboctad & 0x3f
    elif isinstance(suboctad, GCode):
        value = mat24.ploop_cap(gcode, suboctad.value)
        suboctad_ = mat24.cocode_to_suboctad(value, gcode)
    else:
        value = Cocode(suboctad).cocode 
        suboctad_ = mat24.cocode_to_suboctad(value, gcode)
    cocode = mat24.suboctad_to_cocode(suboctad_, gcode)
    if import_pending:
       complete_import()
    result = XLeech2(ploop, cocode)
    subtype =  result.xsubtype
    assert subtype in [0x22, 0x42], (hex(subtype), hex(ploop), hex(cocode), hex(result.value))
    # complent a complement of an octad
    if subtype == 0x42:
        result.value  ^= 0x800000
    return result







"""Configuration for development"""

import sys
import os



# directory names
DEV_DIR = os.path.dirname(os.path.realpath(__file__))
PACKAGE_DIR = os.path.dirname(DEV_DIR)
SRC_DIR = os.path.dirname(PACKAGE_DIR)
ROOT_DIR = os.path.dirname(SRC_DIR)

sys.path.insert(0, ROOT_DIR)
from config import INT_BITS, PRIMES, C_DIR, PXD_DIR
assert sys.path.pop(0) == ROOT_DIR






import sys 
import os
import time
sys.path.insert(0, os.path.abspath("src"))
#print(sys.path)

print("Importing python dlls..")
t_start = time.time()
import mmgroup 
from mmgroup import mat24
from mmgroup import generators
from mmgroup import mm
print("Importing python for specific characteristics..")
from mmgroup.mm_space import characteristics
pp = characteristics()
t = time.time() - t_start
print("done after %.2f seconds." % t)
s = "Representations of the Monster modulo %s are supported." 
print(s % ", ".join(map(str,pp)))




import sys
import os


if __name__ == "__main__":
    sys.path.append("../../../")

from mmgroup.dev.mat24.mat24_ref import Mat24 as mat24
from mmgroup.dev.mat24.make_addition_table import make_addition_operations
from mmgroup.dev.mat24.make_addition_table import check_addition_operations
from mmgroup.dev.mat24.make_addition_table import display_addition_operations

def make_short_ops_table(verbose = 0):
    if verbose:
        s = "Map cocode basis to short cocode vectors [0,i], 1 <= i < 24"
        print(s)
    m = [0] + [mat24.vect_to_cocode(1 + (1 << i)) for i in range(1,24)]
    for i in range(1,24):
        cc = mat24.cocode_syndrome(m[i], 0)
        assert cc == (1 << i) + 1
        if verbose:
            print("0x%03x : 0x%06x" % (m[i], cc))
    singleton = True
    ops = make_addition_operations(m, singleton, granularity = 4)
    check_addition_operations(m, ops, singleton)
    if verbose:
        display_addition_operations(ops)
        print("%d operations" % len([x for x in ops if len(x) >= 4]))
    return ops


def C_table_from_short_ops(ops):
    in_table = []  
    op_table = []
    regs = {}
    for op in ops:
        if len(op) < 4:
            if op[1] > 0:
                assert op[1] == 1 << len(in_table)
                in_table.append(op[0])
                regs[op[1]] = op[0]
        else:
            op_table += [regs[x] for x in op[2:]]
            op_table.append(op[0])
            regs[op[1]] = op[0]
    assert max(op_table) < 24
    assert len(op_table) % 3 == 0
    assert len(ops) == len(in_table) + len(op_table) // 3 + 1
    return in_table, op_table


class ShortCocodeTables:
    ops = make_short_ops_table(verbose = 0)
    in_table, op_table = C_table_from_short_ops(ops)

    tables = {
       "ShortCocode_InTable" : in_table,
       "ShortCocode_OpTable" : op_table,
       "ShortCocode_LenOpTable" : len(op_table),
       "ShortCocode_NumOps" : len(op_table) // 3,
    }

    directives = {}


    @staticmethod
    def display_op():
        make_short_ops_table(verbose = 1)


if __name__ == "__main__":
    ShortCocodeTables.display_op()
    print("Input table:")
    print(ShortCocodeTables.in_table)
    print("Operation table:")
    print(ShortCocodeTables.op_table)
   







"""Implements fast arithmetic of bit vectors

A bit vector b = (b[0],...,b[n-1]) , b[i] in (0,1) is coded as the
integer sum( [ b[i] * 2**i  for i in range(n) ] ). The is a bit 
inconvenient to  read for a human, but with this encoding functions 
operating on bit vectors can be defined and implemented in a much more 
consistent way. Function bin() also allows to print the bits of an 
integer in reverse order.  The first index of a vector or matrix is 
always zero, as usual in Python or C. 

Generally speaking, bit vectors are mainly used as modules with a
group operating on them. Elements of such groups may be matrices
or permutations on bit vectors. Here bit vectors are considered
as row vectors, and bit matrices or permutation always operate
on the right. Thus  x * g1 * g2, where x is a bit vector and g1,
g2 are elements of a group, means that first g1 is applied to v
and then g2 is applied to the bit vector v * g1.

Bit matrices are represented as lists of integers, where the i-th 
element of the list encodes the i-th row vector of the matrix. 
Function bit_mat_mul() multiplies an arbitrary number of
matrices. The first factor may also be a bitvector. In the same 
way as an integer in binary representation can be preceeded by
an arbitrary number of of zero bits (without changing its value),
we assume that a list representing a matrix can be followed by
an arbitrary number of zero entries so that multiplcation of
bit matrices is always possible.

A permutation p is also represented as a list of integers.
Here p[i] = k means that the bit at position i is mapped to
position k. More strictly, we will interpret a permutation p
represented by a list [ p[0], p[1],...,p][n-1] ] with
p[i] = k as a bit matrix b with b[i,j]  = 1 if j==k and
b[i,j] = 0 if j != k. Function  bit_perm_mul() multiplies a
bit vector or a permutation with an arbitrary number of
permutations. Function  bit_perm_2_mat() transforms a
permutation into an equivalent bit matrix.

""" 
from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from random import randint
import types
from operator import __or__ , __xor__
from functools import reduce
from collections.abc import Iterable
from numbers import Integral

import numpy
from numpy import array, zeros, uint8, uint16, uint32, uint64, int64


def isinteger(x):
    """Deprecated, use numbers.Integral() instead!"""
    return  isinstance(x, (Integral, numpy.integer))

def lmap(*args):
    """For compatibility with map() in python 2"""
    return list(map(*args))


def lrange(*args):
    """For compatibility with range() in python 2"""
    return list(range(*args))


def bitlen(x):
    """returns the length of the binary representation of an integer x >= 0"""
    assert x >= 0, "bitlen(x) requires nonegative integer x"
    try:
        return (int(x)).bit_length()
    except:  # for python versions < 2.7 or 3.1:
        l, m, sh = 0, 0x10000, 16
        while x >= m:   
            #x, l = x>>sh, l+sh
            m, sh = m << sh, sh << 1
        #assert m == 1 << sh
        while sh:
            if x >= m: x, l = x>>sh, l+sh
            sh >>= 1
            m = m >> sh  
        while x: x, l = x>>1, l+1
        return l



def hibit(x):
   """return the position of the highest bit of an integer"""
   if x==0: raise ZeroDivisionError("Zero has no highest bit")
   return bitlen(x) - 1





def v2(*args):
   """Return the 2-adic value of an integer or of a set of integers.

   For an integer ``i = o * 2**k``, ``o`` odd, we have ``v2(i) = k``.

   If ``x`` is iterable or a numpy array then ``v2(x)`` is the 
   minimum of the values ``v2(i)`` for all entries of ``i`` of ``x``.

   If several arguments ``x_1,...,x_n`` are given then 
   ``v2(x_1,...,x_n)`` is the minimum of all values ``v2(x_i)``.

   The function raises ZeroDivisionError if all integers occuring as 
   arguments (or contained in an argument) are zero.
   """
   x = 0
   for a in args:
       if isinstance(a, Integral):
           x |= int(a) 
       elif isinstance(a, numpy.ndarray):
           try:
               x |= int(reduce(__or__, a.reshape(-1), 0))
           except TypeError:
               if issubclass(a.dtype.type, np.integer):
                    # workaround for the case a.dtype == numpy.uint64
                    x |= reduce(__or__, map(int,a.reshape(-1)), 0)
               else:
                   raise
       elif isinstance(a, Iterable):
           x |= int(reduce(__or__, a, 0))
       else:
          err = "Function v2() undefined for a %s object"
          raise TypeError(err % type(a))
   l = (x & -x).bit_length() - 1
   if l >= 0:
       return l
   raise ZeroDivisionError("Function v2(x) is undefined for x = 0")



def bits2list(x):
   """returns the list of positions of one-bits of integer x

   List of bit positions is sorted in ascending order
   """
   l = bitlen(x)
   return [ i for i in range(l) if (x >> i) & 1 ]


def list2bits(lst):
   """return number x with bit i of x  set for all i in the list 'lst' """
   return reduce(__or__, [ 1<<i for i in lst ], 0) 





def list_xor(ListList):
   """Given a list of lists, the list of  XOR sums of the cartesian product are returned"""
   if len(ListList) <= 1:
       return ListList[0] if len(ListList) else []
   d = len(ListList) >>1
   l1, l2 = ListXor(ListList[:d]), ListXor(ListList[d:])
   return [x ^ y for x in l1 for y in l2]





def iter_bitweight(k,n):
    """Iterate through all integers 0 <= x < n with bit weight k"""
    if k > 0:
        x = (1 << k) - 1  # smallest integer of bit weight k
        while x < n:
            yield x
            # use Gosper's hack to get next integer x of bit weight k
            c = x & -x
            r = x + c
            x = (((x ^ r) >> 2) // c) | r
    elif n > 0:
        yield 0

def iter_ascending_bitweight(n):
    """Iterate through all integers 0 <= x < n by ascending bit weight"""
    if n > 0:
        yield 0
        k = 1
        while 1 << k <= n:
            for y in iter_bitweight(k,n):
                yield y  
            k = k + 1




def bw24(x):
    """Return the bit weight of the lowest 24 bits of x"""
    x = (x & 0x555555) + ((x & 0xaaaaaa) >> 1)
    x = (x & 0x333333) + ((x & 0xcccccc) >> 2)
    x = (x + (x >> 4)) & 0xf0f0f
    return (x + (x >> 8) + (x >> 16)) & 0x1f 


def bitweight(x):
   """return the bit weight of the  word 'x'. """
   assert x >= 0, "bitweight(x) requires integer x >= 0"
   w = 0
   while x:
        w += bw24(x)
        x >>= 24
   return w 

def bitparity(x):
   """return the bit parity of the  word 'x'. """
   assert x >= 0, "bitparity(x) requires integer x >= 0"
   while x > 0xffffffff:
       x = (x >> 32) ^ (x & 0xffffffff)
   x ^= x >> 16
   x ^= x >> 8
   x ^= x >> 4
   return (0x6996 >> (x & 15)) & 1




def reverse24(x, k=24):
    """reverses order of bits 0,...,k-1 of an integer. Default is k=24.

    If any bits apart form bit 0,...,k-1 are set, an execption is raised.
    """
    if x & -(1 << k):
        raise ValueError("Too high bits are set for bit reversal")
    return sum( ((x >> (k-i-1)) & 1) << i  for  i in range(k) )



def bin(x, l=24, d=8, reverse=False):
   """returns the binary representation of an 'l'-bit integer 'x' as a string

   Leading zeros are printed, blanks are inserted between 'd'-bit blocks.
   'd' may also be a list of bit positions where to insert blanks. If
   'reverse' is True, the string starts with bit 0 (which is fine for bit
   vectors), otherwise it starts with bit l-1 (as usual for integers). 
   """
   s = ""
   if not isinstance(d, list):  d=range(d,l,d)
   if reverse:
      for i in range(l):
         if x & (1<<i): s += '1'
         else: s += '0'
         if i+1 in d: s += ' '
   else:
      for i in range(l-1,-1,-1):
         if x & (1<<i): s += '1'
         else: s += '0'
         if i in d: s += ' '
   return s



def binbv(x, l=24, d=8):
   """shorthand for bin(x, l, d, reverse=True)"""
   return bin(x, l, d, reverse=True)

      


def pivot_binary_high(basis):
   """Pivot over a basis of integers considered as bit vectors

   Here 'basis' list a list of integers corresponding to bit vectors,
   and the elements of the list 'basis' are considered as the rows of a 
   bit matrix defining a basis of a linear space in a vector space 
   over {0,1}.

   The function performs Gaussian elimination to change the basis to 
   a standard form. Here pivoting over bit columns is done, so that
   at the end each pivoted bit column contains exactly one nonzero bit.
   Redundant rows of the original basis are removed.

   Bit columns are not exchanged. The function returns a pair
   (basis, columns), where 'basis' is the modified basis and 'columns'
   is a list of integers containing the index of the pivoted column
   used in the corresponding row of the basis. Here index i corresponds
   to the bit column with valence i.    

   The highest possible bit column is always taken for pivoting.  
   """
   basis, columns = list(basis)[:], []
   for i in range(len(basis)):
       m = max(basis[i:])
       j = basis[i:].index(m) + i
       try:
           mi = hibit(m) 
       except:
           del basis[i:]
           break
       else:
           mask = 1 << mi  
           columns.append(mi)
           basis[i], basis[j] = m, basis[i]
           for j in range(0, len(basis)):
               if j != i and basis[j] & mask: 
                    basis[j] ^= m 
   return basis, columns   







def pivot_binary_low(basis):
   """Pivot over a basis of integers considered as bit vectors

   Here 'basis' list a list of integers corresponding to bit vectors,
   and the elements of the list 'basis' are considered as the rows of a 
   bit matrix defining a basis of a linear space in a vector space 
   over {0,1}.

   The function performs Gaussian elimination to change the basis to 
   a standard form. Here pivoting over bit columns is done, so that
   at the end each pivoted bit column contains exactly one nonzero bit.
   Redundant rows of the original basis are removed.

   Bit columns are not exchanged. The function returns a pair
   (basis, columns), where 'basis' is the modified basis and 'columns'
   is a list of integers containing the index of the pivoted column
   used in the corresponding row of the basis. Here index i corresponds
   to the bit column with valence i.    

   The lowest possible bit column is always taken for pivoting.  
   """
   basis, columns = basis[:], []
   for i in range(len(basis)):
       m = reduce(__or__, basis[i:], 0)
       try:
           mi = v2(m) 
       except:
           del basis[i:]
           break
       else:
           mask = 1 << mi  
           j = [x & mask for x in basis[i:]].index(mask) + i
           columns.append(mi)
           m = basis[j]
           basis[i], basis[j] = m, basis[i]
           for j in range(0, len(basis)):
               if j != i and basis[j] & mask: 
                   basis[j] ^= m
   return basis, columns   









_maxbits = {uint8:8, uint16:16, uint32:32}

def lin_table(lst, dtype = uint32, t0 = 0):
    """return a linear table t with t[0]=t0, t[2**i]=lst(i).

    For all entries i,j of t we have t[i^j] ^ t[0] = t[i] ^ t[i].
    t[0] is set to the value t0 (default is 0).
    
    The table ist returned as a numpy array of type 'dtype',
    with dtype = uint8, unit16 or uint 32.
    """
    if not 0 <= (reduce(__or__,lst,0) | t0) <= 1 << _maxbits[dtype]:
        print( hex(reduce(__or__,lst,0) | t0) )
        raise ValueError("entries too large for table")
    a = numpy.zeros(1 << len(lst), dtype = dtype)
    a[0] = t0
    for i, x in enumerate(lst):
        a[1<<i:2<<i].fill(x)
        a[1<<i:2<<i] ^= a[0:1<<i]
    return a


def bit_mat_iszero(a):
    """return 0 if a is a zero matrix"""
    if isinteger(a):
         return a == 0
    return sum(lmap(abs,a)) == 0

def bit_mat_mul(a, *args):
    """ Multiplies 'a' with an arbitrary number of bit matrices.

    Parameter 'a' may be a bit vector or a list representing a bit matrix.
    Representation of bit vectors and matrices see comment on module.
    The result returned has the same shape as parameter 'a'. 
    """
    if len(args) == 1:
        b = args[0]
        if isinteger(a):
            return reduce(
                   __xor__, [x for i,x in enumerate(b) if a & (1<<i)], 0 )
        else:
            return [ reduce(
                  __xor__, [x for i,x in enumerate(b) if y & (1<<i)], 0 )
               for y in a ]
    if len(args) == 0:
        return a
    return  bit_mat_mul(bit_mat_mul(a, args[0]), *args[1:])



def bit_mat_transpose(a, ncolumns=None):
    """return the  transpose of bit Matrix a.
 
    Optionally, the number of columns of a may be given.

    """
    ncols = bitlen(reduce(__or__, a, 0))
    if ncolumns is None: 
        ncolumns = ncols 
    elif ncols > ncolumns:
         raise ValueError( "Bad entry in bit matrix" )
    t = [0] * ncolumns
    for i, l in enumerate(a):
        for j in range(ncolumns):
            if (l >> j) & 1:
                 t[j] ^= 1 << i
    return t





def bit_mat_inverse(a):
    """return the inverse of bit Matrix a."""
    msg = "Bit matrix is not invertible"
    ncols = bitlen(reduce(__or__, a, 0))
    if ncols != len(a):
         raise ZeroDivisionError( msg )
    perm = [None] *  ncols
    hicol = 1 << ncols
    ah = [ x + (hicol << i) for i,x in enumerate(a) ]
    for i in range(ncols):
        piv = v2(ah[i])
        if piv >= ncols:
            raise ZeroDivisionError( msg )
        perm[piv] = i
        msk = 1 << piv
        for j in range(ncols):
            if ah[j] & msk and i != j:
                ah[j] ^= ah[i]
    return [ ah[perm[i]] >> ncols  for i in range(ncols) ]


 

def bit_mat_det(a):
    """return the determinant of bit Matrix a."""
    try:
        bit_mat_inverse(a)
        return 1
    except ZeroDivisionError:
        return 0


def bit_mat_orthogonal_complement(a, ncolumns=None):
    """return an orthognal complement of bit Matrix a.

    Optionally, the number of columns of a may be given.
    Here matrix 'a' is considered a matrix of row vectors generating a
    linear subspace W of V = GF(2)**ncolumns. The function returns a
    matrix of row vetors generating the orthognal complement of W with
    respect to the standard Euclidean inner product in V.      
    """
    # This can be improved using the ideas in bit_mat_inverse !!
    if ncolumns == None:
         ncolumns = bitlen(reduce(__or__, a, 0))
    piv_a, piv_cols = pivot_binary_low(a)
    for i in range(ncolumns):
        if not i in piv_cols:
            piv_a.append(1 << i)
    ao = bit_mat_transpose(bit_mat_inverse(piv_a), ncolumns)
    return ao[len(piv_cols):]


def bit_mat_rank(a):
    """return the rank of a bit marix a"""
    return len(pivot_binary_high(a)[0])


def bit_mat_random(nrows, ncolumns):
    """return a random bit matrix with given number of rows and columns"""
    a, rmax = [],  ( 1 << ncolumns) - 1
    for i in range(nrows):
         a.append(randint(0, rmax))
    return a


def bit_mat_basis_spanned(basis, dtype = None):
    """return an array of the bitvectors spanned by a basis

    The function returns an array A of length 2**len(basis) containing
    the bit vectors spanned by the basis.

    Entry A[sum(b_i * 2*i)] is sum(b_i * basis[i]), b[i] = 0,1.

    If dtype is not None, a numpy array of the corresponding dtype
    is returned.
    """
    sp = [0]
    for x in basis:
        sp += [ y ^ x for y in sp ] 
    if not dtype is None:
        sp = numpy.array(sp, dtype = dtype)
    return sp



      

                        

def bit_perm_mul(a, *args):
    """ Multiplies 'a' with an arbitrary number of bit permutations.

    Parameter 'a'  may be a bit vector or a list representing a permutation.
    Representation of bit vectors and permutations see comment on module.
    The result returned has the same shape as parameter 'a'. 
    """
    if len(args) == 1:
        b = args[0]
        if isinteger(a):
            return reduce(
                   __xor__, [1<<x for i,x in enumerate(b) if a & (1<<i)], 0 )
        else:
            return [ b[x]  for  x in a ]
    if len(args) == 0:
        return a
    return  bit_perm_mul(a, *(args[:-2] + ([args[-1][x] for x in args[-2] ],)))  
 
   
    raise NotImplementedError("")


def bit_perm_2_mat(a):
    """transforms a permutation into a bit matrix.

    Representation of permutations and bit matrices see comment 
    on module.  
    """
    if isinteger(a): return a
    return [1 << x for x in a]



def rand_perm(n):
    """returns a list representing a random permutation of n elements"""
    if n == 1: return [0]
    else:
       a = randint(0, n-1)
       p1 = [x for x in range(a,n)+range(a)]
       p2 = [n-1] + rand_perm(n-1)
       return bit_perm_mul(p1,p2)
        








r"""Module ``build_ext_steps`` provides a customized version of 
the 'build_ext' command for setup.py.

It should be placed in the same directory as the module ``setup.py``.

Distributing python packages
............................

The standard toolkit for distributing python packages is the 
``setuptools`` package. Here the user types::

   python setup.py build_ext

at the console for building the extensions to the python package, which 
are typically written in a language like C or C++ for the sake of speed. 
We may use e.g. the ``Cython`` package to write python wrappers for the 
functions written in C or C++. The ``setuptools`` package supports the
integration of ``Cython`` extensions.

The ``setup.py`` script describes a list of extensions, where each
extension is an instance of class ``Extension`` which is provided  
by  ``setuptools``. Then it calls the ``setup`` function which builds
all these extensions:

.. code-block:: python

  from setuptools import setup
  from setuptools.extension import Extension   

  ext_modules = [
      Extension(
          ...  # description of first extension
      ),
      Extension(
          ...  # description of second extension
      ),
      ...
  ]

  setup(
      ..., # Standard arguments for the setup function
      ext_modules = ext_modules,   # List of extensions to be built
  )

We assume that the reader is familiar with the standard python setup 
process. For background, we refer to


https://setuptools.readthedocs.io/en/latest/


A new paradigm for building python packages
...........................................


This ``build_ext_steps`` module supports a new paradigm for building a 
python extension:

  * A python program ``make_stage1.py`` creates a C program 
    ``stage1.c``.

  * We create a python extension ``stage1.so`` (or ``stage1.pyd`` 
    in Windows) that makes the functionality of ``stage1.c``
    available in python.

  * A python program ``make_stage2.py`` creates a C program 
    ``stage2.c``. Here ``make_stage2.py`` may import ``stage1.so`` 
    (or ``stage1.pyd``).

  * We create a python extension  that makes the functionality of 
    ``stage2.c``  available in python.

  * etc.

This paradigm is not supported by the ``setuptools`` package.

Using class ``BuildExtCmd`` for the new paradigm
................................................

For using the new building paradigm we have to replace the standard 
class ``build_ext`` by the class ``build_ext_steps.BuildExtCmd``. 

.. code-block:: python

  from setuptools import setup
  from build_ext_steps import Extension   
  from build_ext_steps import BuildExtCmd

  ext_modules = [
      # description of extension as above
  ]

  setup(
      ..., # Standard arguments for the setup function
      ext_modules = ext_modules,   # List of extensions to be built
      cmdclass={
         'build_ext': BuildExtCmd, # replace class for build_ext
      },
  )

This change has a few consequences:

  * It is guaranteed that the extension are build in the given order

  * Extensions are always build in place (option ``build_ext --inplace``)
    (The current version does not support building the extension in
    a special build directory.)

  * The building of all extensions is now forced 
    (option ``build_ext --f``), regardless of any time stamps.

  * A keyword argument ``extra_compile_args`` and ``extra_link_args`` 
    for  an instance of class ``Extension`` may be a dictionary

       'compiler' : <List if arguments>

    instead of a list of arguments. Here 'compiler' is a string
    describing a compiler. For a list of compilers, run

      ``python setup.py build_ext --help-compiler`` .


Apart from these changes, an extension is created in the same way 
as with ``setuptools``.

For a documentation of the ``Extension`` class in the ``setuptools``
package, see

https://docs.python.org/3/distutils/apiref.html?highlight=extension#distutils.core.Extension


Inserting user-defined functions into the build process
.......................................................


Module ``build_ext_steps`` provides a class ``CustomBuildStep``
for adding user-defined functions to the build process.

In the list ``ext_modules`` of extensions, instances of class 
``CustomBuildStep`` may be mixed with instances of class ``Extension``,
Class ``CustomBuildStep`` models an arbitrary  sequence of functions to 
be executed.

The constructor for that class takes a string 'name' describing the
action of these functions, followed by an arbitrary number of lists, 
where each list describes a function to be executed.

Here the first entry of each list is either a string or a callable
python function. If the first entry is a string then a subprocess
with that name is called. Otherwise the corresponding python function
is executed. Subsequent entries in the list are arguments given to
the subprocess or to the function.

Keyword arguments in the constructor are passed to function
``subprocess.check_call`` when executing a subprocess and 
ignored when executing a python function.

Such a subprocess may be e.g. a step that generates C code to be
used for building a subsequent python extension.    

Its recommended to use the string ``sys.executable`` (provided
by the ``sys`` package) instead of the string ``'python'`` for 
starting a python subprocess.


Building shared libraries
.........................

Module ``build_ext_steps`` provides another class ``SharedExtension``
for building  a shared library (or a DLL in Windows).

In the list ``ext_modules`` of extensions, instances of class 
``SharedExtension`` may be mixed with instances of other  classes.

Arguments for the constructor of class ``SharedExtension`` are the
same as for class ``Extension``. Especially, the following keyword 
arguments are recognized:


  * ``name``, ``sources``, ``include_dirs``, ``libraries``,
    ``define_macros``, ``undef_macros``

Here ``name`` is a Python dotted  name without any extension as in
class ``Extension``. The command uses the C compiler to build
the shared library and stores it at the location given by ``name``
in the same way as a python extension is built. The appropriate
extension (e.g. '.so' for unix and '.dll' for Windows) is 
automatically appended to the file name of the shared library.

The user should provide an additional keyword argument ``implib_dir`` 
specifying a directory where to store the import library for a Windows 
DLL. In Windows, the import library for ``foo.dll`` has the name 
``libfoo.lib``. If a program uses a Windows DLL then it should be 
linked to that import library. In unix operating systems there is no 
concept similar to an import library.

Caution!!

The current class ``SharedExtension`` supports Windows DLLs containing
C programs (no C++) compiled with the ``mingw32`` or the ``msvc``
compiler only.


Using a shared library in an extension
......................................

The reason for building a shared library is that several python
extensions may use the same shared library.

The way how a shared library (or a Windows DLL) is linked to a program
using that library depends on the operating system.

After creation, an object of class  ``SharedExtension`` contains an
attribute ``lib_name``. Attribute ``lib_name`` contains name of the 
library (or of the import library in Windows) that has to be linked
to a program using the shared library.

The user may have to perform some os-specific steps for making the 
library available for python extension. Therefore he may read
the variable ``os.name`` (which has value ``'nt'`` for Windows and
``'posix'`` for unix) and use class ``CustomBuildStep`` for performing
the appropriate steps.

For Windows one has to add the directory of the import library
(discussed in the previous section) to the search path for the library
by specifying that directory in the ``'library_dirs'`` keyword 
argument for class Extension. Also, one has to specify the name
of the import library in the  ``'libraries'`` keyword argument for 
class Extension. This is sufficient if the python extension and
the DLLs used by that extension are in the same directory.

More involved cases of shared libraries and the procedures required
for other operating systems are out of the scope of this documentation.

Using an extension in a subsequent build step
.............................................

Once a python extension has been built, it can also be used in
a subsequent step of the build process, e.g. for calculating large
arrays of constants for C programs.

This approach works well on a Windows system, but it might not work
on other operating systems. Here it is a good idea to write a
pure-python substitute for any C extension to be used in a subsequent
build step. This may slow down the build process considerably. But it 
is better to have a slow build process than no build process at all.


Technical remarks about shared libraries
........................................

The functionality for building a Windows DLL with the ``migw32`` 
compiler in coded in function ``make_dll_win32_gcc``. One could have 
used the functionality of class  ``distutils.ccompiler`` instead, 
but the author has decided not to dive any deeper into the source 
code of the ``distutils`` package. It may be worth using class  
``distutils.ccompiler`` for porting the functionality of classes 
``BuildExtCmd`` and  ``SharedExtension`` to other compilers and 
operating systems.

The ``mingw32`` compiler is not the standard compiler for python on
Windows. C++ files should be compiled with the standard compiler
(which is ``msvc`` for Windows) in order to avoid trouble with 
name mangling.
"""

import sys
import os
import subprocess
from subprocess import CalledProcessError

from glob import glob

import setuptools
from setuptools.extension import Extension as _Extension


from Cython.Distutils import build_ext as _build_ext


class DistutilsError (Exception):
    """The root of all Distutils evil.

    Specifically, I do not know how and when exactly 'distutils'
    will eventually become deprecated.

    Since I'm not willing to dig any deeper into the details of
    setuptools/distutils, I simply copy the relevant exceptions
    here. (This is a weird way to deal with a weird concept!)
    """
    pass

class DistutilsPlatformError (DistutilsError):
    """We don't know how to do something on the current platform (but
    we do know how to do it on some platform) -- eg. trying to compile
    C files on a platform not supported by a CCompiler subclass."""
    pass


class DistutilsSetupError (DistutilsError):
    """For errors that can be definitely blamed on the setup script,
    such as invalid keyword arguments to 'setup()'."""
    pass



_compiler = None


# Class ``Extension`` is equal to class ``Cython.Distutils.build_ext``.
# This may change in future versions.
Extension = _Extension

class CustomBuildStep(_Extension):
    """Model a custom build step as described in the header of this module 
    """
    def __init__(self, name = "Customer_step", *args, **kwds):
        # Set attribute name (for displaying the corresponding build step)
        self.name = name
        if not isinstance(name, str):
            raise AssertionError("'name' must be a string")
        # Set the list of functions
        self.function_list = list(args) or []
        # Set the dict of keywords
        self.keyword_dict = kwds
        # Mark this as a nonstandard instance of class Extension
        self.is_non_standard = True
        # Set all other attributes of class Extension to default values
        self.sources = []
        self.include_dirs = []
        self.define_macros =  []
        self.undef_macros = []
        self.library_dirs =  []
        self.libraries =  []
        self.runtime_library_dirs =  []
        self.extra_objects = []
        self.extra_compile_args = []
        self.extra_link_args =  []
        self.export_symbols =  []
        self.swig_opts =  []
        self.depends =  []
        self.language = None
        self.optional = None
          


class SharedExtension(_Extension):
    """Model a shared library as described in the header of this module 
    """
    def __init__(self,  *args, **kwds):
        self.implib_dir = None
        # If keyword argument "implib_dir" is given, set self.implib_dir
        # to the corresponding value
        if "implib_dir" in kwds:
            self.implib_dir = kwds["implib_dir"]
            del kwds["implib_dir"]
        # Treat all other arguments as in the base class 'Extension'.
        super(SharedExtension, self).__init__(*args, **kwds)
        self.is_non_standard = True
        self.lib_name = self.name.split(".")[-1]
        if os.name == "nt":
            self.lib_name = "lib" + self.lib_name






class  BuildExtCmd(_build_ext):
    """Substitute for class setuptools.extension import Extension

    The ``run`` function of this class treats an instance  of class
    Extension in the usual way, i.e. it builds a python extension. 

    Instances of class ``CustomBuildStep`` or ``SharedExtension``
    are treated as in the description of this module.
    """
    user_options = _build_ext.user_options + [
        ("nprocesses=", None, 
        "Number of prcesses used by class ParallelSteps"
        )
    ]

    def initialize_options(self):
        super(BuildExtCmd, self).initialize_options()
        self.nprocesses = 1

    def finalize_options(self):
        super(BuildExtCmd, self).finalize_options()
        self.nprocesses = max(1, int(self.nprocesses))
   

    def run(self):
        """Implementation of the run() function.

        See class distutils.cmd.Command for background.
        """
        # At this point, 'self.compiler' is None or a string,  but this 
        # will change while running the corresponding distutils command
        self.compiler_name = self.compiler
        if self.compiler_name is None:
            # Then try to get the default compiler name
            from distutils.ccompiler import get_default_compiler
            self.compiler_name = get_default_compiler()

        # Now process all extensions in the given order
        for ext in self.extensions[:]:
            if isinstance(ext, CustomBuildStep):
                # Then we execute a sequence of subprocesses or functions 
                print("\nExecuting custom build step '%s'" %  ext.name)
                for args in ext.function_list:
                    if isinstance(args[0], str):
                        # Then execute a subprocess
                        print(" ".join(map(str, args)))
                        try:
                            subprocess.check_call(list(map(str, args)), 
                                **ext.keyword_dict)
                        except CalledProcessError:
                            err = "\nSubprocess '%s' has failed!"
                            print(err % str(args[0]))
                            raise             
                    else:
                        # Then execute a python function
                        args[0](*(args[1:])) 
            elif isinstance(ext, SharedExtension):
                # Then we build a shared librry using method
                # self.build_shared_extension()

                # Evaluate extra_compile_args and extra_link_args to a string
                ext.extra_compile_args = self.eval_extra_args(
                    ext.extra_compile_args)
                ext.extra_link_args = self.eval_extra_args(
                    ext.extra_link_args)
                print("\nBuilding shared library '%s'" %  ext.name)
                self.build_shared_extension(ext)
            elif isinstance(ext, str):
                # A string is taken as a comment
                pass 
            else:
                # A standard Extension is build in the same way as in the
                # base class. We rely on the fact that setuptools always 
                # builds the python extension in the build directory,
                # even if the 'inplace' option has not been set.

                # Evaluate extra_compile_args and extra_link_args to a string
                ext.extra_compile_args = self.eval_extra_args(
                    ext.extra_compile_args)
                ext.extra_link_args = self.eval_extra_args(
                    ext.extra_link_args)
                # Save the old value of self.compiler because the self.run()
                # method appears to change it so that it self.run() method
                # cannot be applied twice.
                compiler = self.compiler
                extensions = self.extensions
                self.extensions = [ext] # Build one extension only
                inplace = self.inplace
                self.inplace = True     # Always build inplace
                force = self.force      
                self.force = True       # Always force building
                # Run the corresponding mathod of the base class
                super(BuildExtCmd, self).run() 
                # Restore old attributes
                self.extensions = extensions
                self.inplace = inplace
                self.force = True
                self.compiler = compiler


    def eval_extra_args(self, extra_args):
        """Evaluate the dictionary ``extra_args`` to a string.

        Here ``extra_args`` is one of the keyword arguments 
        ``extra_compile_args`` or ``extra_link_args`` passed
        to the constructor of class Extension. The  dictionary
        has entries:

            'compiler' : <List if arguments>

        If ``self.compiler_name`` matches a key 'compiler'  then  
        the list ``extra_args[self.compiler_name]`` is returned. 
        If ``self.compiler_name`` is a string then that string is
        returned.
        
        For a list of compilers, run

            ``python setup.py build_ext --help-compiler`` .

        """
        if isinstance(extra_args, dict):
            return extra_args[self.compiler_name]
        return extra_args
 
    def build_shared_extension(self, ext):
        """Build a shared library

        The shared libarary is described by the instance ``ext``
        of class SharedExtension.
        """ 
        compiler = self.compiler_name
        if os.name == "posix":
            if compiler == 'unix':
                make_so_posix_gcc(self, ext)
            else:
                raise DistutilsPlatformError(
                  "I don't know how to build a posix shared library "
                  "with the '%s' compiler" % compiler)
        elif os.name == "nt":
            if compiler == 'mingw32':
                make_dll_nt_mingw32(self, ext)
            elif compiler == 'msvc':
                make_dll_nt_msvc(self, ext)
            else:
                raise DistutilsPlatformError(
                  "I don't know how to build a Windows DLL "
                  "with the '%s' compiler" % compiler)
        else:
            raise DistutilsPlatformError(
                  "I don't know how to build a shared library "
                  "on platform '%s'" % os.name)

    @classmethod
    def get_shared_ext_filename(cls, ext_name):
        """Convert a python module to the name of a shared library

        Here ``ext_name`` is the   Python dotted  name without any 
        extension as in class ``Extension``.

        The function returns the full path name of the corresponding
        shared library.
        """
        if os.name == "posix":
            suffix = ".so"
        elif os.name == "nt":
            suffix = ".dll"
        else:
            raise DistutilsPlatformError(
                  "I don't know the suffix for a shared library "
                  "on platform '%s'" % os.name)
        ext_path = ext_name.split('.')
        return os.path.join(*ext_path) + suffix

    def get_outputs(self):
        """This extends the corresponding method of class Extension

        It returns a list of output file names. Apart from the names 
        of the created python extensions, that list also contains 
        the names of the created shared libraries.
        """
        extensions = self.extensions
        self.extension = [ x for  x in extensions
            if not getattr(x, "is_non_standard", None)
        ]
        self.extensions = extensions
        standard_outputs = super(BuildExtCmd, self).get_outputs()
        return standard_outputs + self.get_non_standard_outputs()

    def get_non_standard_outputs(self):
        """Auxilary method for self. get_outputs()

        Returns list of the shared libraries (to be) built.
        """
        outputs = []
        for ext in self.extensions[:]:
            if isinstance(ext, SharedExtension):
                outputs.append(self.get_shared_ext_filename(ext.name))
        return outputs

    def get_build_directory(self):
        """Returns name of directory for the output to be built
        """ 
        print("build_dir is", self.build_lib)
        return self.build_lib

    def get_package_dir(self):
        """Return the package directory of the build command"""
        build_py = self.get_finalized_command('build_py')
        package_dir = os.path.abspath(build_py.get_package_dir('.'))
        #print(package_dir)
        return package_dir

###########################################################################################



def make_dll_nt_msvc(cmd, ext):
    """Create a Windows DLL with the mingw compiler"""
    compile_args = [ "/c",  "/W4", "/DMS_WIN64"] + ext.extra_compile_args 
    for ipath in ext.include_dirs:
        compile_args.append("/I " + os.path.realpath(ipath))
    link_args = ["/DLL"] + ext.extra_link_args 
    objects = []

    # add define macros
    for name, value in ext.define_macros:
        if value is None:
            compile_args.append("-D" + name)
        else:
            compile_args.append("-D" + name + "#" + value)

    # add undef macros
    for name in ext.undef_macros:
        compile_args.append("-U" + name)

    # Compile sources and add objects to list 'objects'
    for source in ext.sources:
        args = compile_args[:]
        args.append(os.path.realpath(source))
        obj = os.path.realpath(os.path.splitext(source)[0] + ".obj")
        args.append('/Fo"%s"' % obj)
        objects.append(obj)
        print("cl " + " ".join(args))
        subprocess.check_call("cl " + " ".join(args)) 

    # Link
    largs = link_args[:] + objects 
    for inc_dir in ext.library_dirs:
        # same search path for include files an libraries
        largs.append("/LIBPATH:" + inc_dir)
    for library in ext.libraries: 
        largs.append( "/DEFAULTLIB:" + library )  
    dll_name = cmd.get_shared_ext_filename(ext.name)  
    dll_path =   os.path.join(cmd.get_package_dir(), dll_name)
    largs +=  ["/OUT:" + dll_path ]
    if not ext.implib_dir:
        raise DistutilsSetupError(
                  "Class SharedExtension requires keyword "
                  "argument 'implib_dir' for building Windows DLL" )
    pure_dll_name = ext.name.split(".")[-1]
    #def_path =   os.path.join(ext.implib_dir, "%s.def" % pure_dll_name)
    #largs += [ "-Wl,--output-def," + def_path ]
    implib_path = os.path.join(ext.implib_dir, "lib%s.lib" % pure_dll_name)
    largs += [ "/IMPLIB:%s" % os.path.abspath(implib_path) ]
    print("link " + " ".join(largs))
    subprocess.check_call("link " + " ".join(largs)) 

    #print("removing object files...")
    #for obj in objects:
    #    os.remove(obj)
    print(dll_path + "\nhas been generated.\n")

    #Todo: copy dll to build path if requested
    





def make_dll_nt_mingw32(cmd, ext):
    """Create a Windows DLL with the mingw compiler"""
    compile_args = [ "-c", "-Wall", "-DMS_WIN64"] + ext.extra_compile_args 
    for ipath in ext.include_dirs:
        compile_args.append("-I" + os.path.realpath(ipath))
    link_args = ["-shared",  "-Wall", "-DMS_WIN64"] + ext.extra_link_args 
    objects = []

    # add define macros
    for name, value in ext.define_macros:
        if value is None:
            compile_args.append("-D" + name)
        else:
            compile_args.append("-D" + name + "=" + value)

    # add undef macros
    for name in ext.undef_macros:
        compile_args.append("-U" + name)

    # Compile sources and add objects to list 'objects'
    for source in ext.sources:
        args = compile_args[:]
        args.append(os.path.realpath(source))
        args.append("-o")
        obj = os.path.realpath(os.path.splitext(source)[0] + ".o")
        args.append(obj)
        objects.append(obj)
        print("gcc " + " ".join(args))
        subprocess.check_call("gcc " + " ".join(args)) 

    # Link
    largs = link_args[:] + objects 
    for inc_dir in ext.library_dirs:
        # same search path for include files an libraries
        largs.append("-L" + inc_dir)
    for library in ext.libraries: 
        largs.append("-l" + library )
    dll_name = cmd.get_shared_ext_filename(ext.name)  
    dll_path =   os.path.join(cmd.get_package_dir(), dll_name)
    largs +=  ["-o",  dll_path ]
    if not ext.implib_dir:
        raise DistutilsSetupError(
                  "Class SharedExtension requires keyword "
                  "argument 'implib_dir' for building Windows DLL" )
    pure_dll_name = ext.name.split(".")[-1]
    def_path =   os.path.join(ext.implib_dir, "%s.def" % pure_dll_name)
    largs += [ "-Wl,--output-def," + def_path ]
    implib_path = os.path.join(ext.implib_dir, "lib%s.lib" % pure_dll_name)
    largs += [ "-Wl,--out-implib," + implib_path ]
    print("gcc " + " ".join(largs))
    subprocess.check_call("gcc " + " ".join(largs)) 

    print("removing object files...")
    for obj in objects:
        os.remove(obj)
    print(dll_path + "\nhas been generated.\n")

    #Todo: copy dll to build path if requested
    


def make_so_posix_gcc(cmd, ext):
    """Create a posix shared library with gcc"""
    compile_args = [ "-c", "-Wall"] + ext.extra_compile_args 
    for ipath in ext.include_dirs:
        compile_args.append("-I" + os.path.realpath(ipath))
    link_args = ["-shared",  "-Wall"] + ext.extra_link_args 
    objects = []

    # add define macros
    for name, value in ext.define_macros:
        if value is None:
            compile_args.append("-D" + name)
        else:
            compile_args.append("-D" + name + "=" + value)

    # add undef macros
    for name in ext.undef_macros:
        compile_args.append("-U" + name)

    # Compile sources and add objects to list 'objects'
    for source in ext.sources:
        args = compile_args[:] + ["-fPIC"]
        args.append(os.path.realpath(source))
        args.append("-o")
        obj = os.path.realpath(os.path.splitext(source)[0] + ".o")
        args.append(obj)
        objects.append(obj)
        print("cc " + " ".join(args))
        subprocess.check_call(["cc"] + args) 

    # Link
    largs = link_args[:] + objects 
    for inc_dir in ext.library_dirs:
        # same search path for include files an libraries
        largs.append("-L" + inc_dir)
    for library in ext.libraries: 
        largs.append("-l" + library )
    dll_name = cmd.get_shared_ext_filename(ext.name).split("/")
    dll_name[-1] =  "lib" + dll_name[-1]  
    dll_path =   os.path.join(cmd.get_package_dir(), *dll_name)
    largs +=  ["-o",  dll_path ]
    print("cc " + " ".join(largs))
    subprocess.check_call(["cc"] + largs) 

    print("removing object files...")
    for obj in objects:
        os.remove(obj)
    print(dll_path + "\nhas been generated.\n")

    #Todo: copy dll to build path if requested
    







